import os.path
from typing import List, Tuple

import cv2
import numpy
import pytest
import skimage

FIXTURE_DIR: str = '/Users/rodrigo/src/pyjamas_dev/pyjamas/tests/unit/fixtures/'
LOGO_FIXTURE: str = 'original_logo.tif'
IMAGE_FIXTURE: str = 'image.tif'
CLASSIFIER_FIXTURE: Tuple[str, str] = ('classifier_lr.cfr', 'classifier_svm.cfr')
DISPLAY_FIXTURE: str = 'display.tif'
BINMASKS_FIXTURE: str = 'binmasks.tif'
MOVIE_FIXTURE: str = 'movie.avi'
PJSANNOTATIONS_FIXTURE: str = 'annotations.pjs'
LOGOANNOTATIONS_FIXTURE: str = 'logoannotations.pjs'
REGISTRATIONANNOTATIONS_FIXTURE: str = 'logoregannotations.pjs'
MATANNOTATIONS_FIXTURE: str = 'annotations.mat'
SINGLE_CELL_PJSANNOTATIONS_FIXTURE: str = 'cell_00001.pjs'
PROJECTION_FIXTURE: Tuple[str, str] = ('max_projection.tif', 'sum_projection.tif')
ZPROJECTIONS_FIXTURE: Tuple[str, str] = ('xz.tif', 'yz.tif')
ROTATED_CW_FIXTURE: str = 'rotated_cw.tif'
FLIPPED_LR_FIXTURE: str = 'flipped_lr.tif'
FLIPPED_UD_FIXTURE: str = 'flipped_ud.tif'
INVERTED_FIXTURE: str = 'inverted.tif'
GRADIENT_FIXTURE: str = 'gradient.tif'
KYMOGRAPH_FIXTURE: str = 'kymograph.tif'
REGISTERED_FIXTURE: str = 'registered.tif'
TMP_IMG_FILE: str = 'image.tif'
TMP_DIR: str = '/Users/rodrigo/src/pyjamas_dev/pyjamas/tests/tmp/'
TMP_DISP_FILE: str = 'display.tif'
TMP_BINMASKS_FILE: str = 'binmasks.tif'
TMP_MOVIE_FILE: str = 'movie.avi'
TMP_PJS_FILE: str = 'annotations.pjs'
TMP_MAT_FILE: str = 'annotations.mat'
TMP_SINGLE_CELL_PJS_FILE: str = 'cell_00001.pjs'
PIX_PERCENTILE_LIMITS: Tuple[int, int] = (10, 90)
GAUSSIAN_SIGMA: float = 3.0
RESCALE_TUPLE: Tuple[float, float] = (3.0, 0.5)
FINDSEEDS_SIGMA: float = 2.0
FINDSEEDS_WINDOWSIZE: int = 16
FINDSEEDS_MINDIST: float = 3.0
FINDSEEDS_BINCLOSINGS: int = 2
FINDSEEDS_ANNOTATIONSPATH: str = 'fiducials_sigma2_window16_closings2_mindist3.pjs'
PROPAGATESEEDS_ANNOTATIONSPATH: str = 'propagatedfiducials_sigma2_window16_closings2_mindist3.pjs'
PROPAGATESEEDS0_ANNOTATIONSPATH: str = 'propagatedfiducials0_sigma2_window16_closings2_mindist3.pjs'
EXPANDSEEDS_ANNOTATIONSPATH: str = 'expandedfiducials_sigma2_window16_closings2_mindist3.pjs'
EXPANDNPROPAGATESEEDS_ANNOTATIONSPATH: str = 'expandedpropagatedfiducials_sigma2_window16_closings2_mindist3.pjs'
PROJECT_CONCAT_DIR: str = 'projcat'
TMP_PROJCAT_FILE: str = 'projcat.tif'
BATCH_MEASURE_DIR: str = 'batch_measure/results'
BATCH_FLATFIELD_DIR: str = 'batch_ffc/images_for_correction'
BATCH_FLATFIELD_OUTPUTDIR: str = 'batch_ffc/corrected_images'
BATCH_FLATFIELD_FLATFIELDIMAGEPATH: str = 'batch_ffc/ffc_image.tif'
BATCH_FLATFIELD_DARKIMAGEPATH: str = 'batch_ffc/dark_field_image.tif'
BATCH_FLATFIELD_SUBSTR: str = ''
BATCH_FLATFIELD_FILESUFFIX: str = '_ffc_corrected'
BATCH_FLATFIELD_CROPDIMS: Tuple[int, int] = (512, 512)
@pytest.fixture
def image_fixture():
    return skimage.io.imread(os.path.join(FIXTURE_DIR, IMAGE_FIXTURE))


@pytest.fixture
def display_fixture():
    return skimage.io.imread(os.path.join(FIXTURE_DIR, DISPLAY_FIXTURE))


@pytest.fixture
def binmasks_fixture():
    return skimage.io.imread(os.path.join(FIXTURE_DIR, BINMASKS_FIXTURE))


@pytest.fixture
def pjsannotationspath_fixture():
    return os.path.join(FIXTURE_DIR, PJSANNOTATIONS_FIXTURE)


@pytest.fixture
def logoannotationspath_fixture():
    return os.path.join(FIXTURE_DIR, LOGOANNOTATIONS_FIXTURE)


@pytest.fixture
def singlecellpjsannotationspath_fixture():
    return os.path.join(FIXTURE_DIR, SINGLE_CELL_PJSANNOTATIONS_FIXTURE)


@pytest.fixture
def matannotationspath_fixture():
    return os.path.join(FIXTURE_DIR, MATANNOTATIONS_FIXTURE)


@pytest.fixture
def registrationannotationspath_fixture():
    return os.path.join(FIXTURE_DIR, REGISTRATIONANNOTATIONS_FIXTURE)


@pytest.fixture
def imagepath_fixture():
    return os.path.join(FIXTURE_DIR, IMAGE_FIXTURE)


@pytest.fixture
def classifierpath_fixture():
    return (os.path.join(FIXTURE_DIR, classifier_fixture_filename) for classifier_fixture_filename in
            CLASSIFIER_FIXTURE)


@pytest.fixture
def movie_fixture():
    movie = cv2.VideoCapture(os.path.join(FIXTURE_DIR, MOVIE_FIXTURE))

    success, movie_array = movie.read()
    success, theslice = movie.read()

    while success:
        movie_array = numpy.concatenate((movie_array, theslice), axis=0)
        success, theslice = movie.read()

    return movie_array


@pytest.fixture
def rotatecwpath_fixture():
    return os.path.join(FIXTURE_DIR, ROTATED_CW_FIXTURE)


@pytest.fixture
def fliplrpath_fixture():
    return os.path.join(FIXTURE_DIR, FLIPPED_LR_FIXTURE)


@pytest.fixture
def flipudpath_fixture():
    return os.path.join(FIXTURE_DIR, FLIPPED_UD_FIXTURE)


@pytest.fixture
def invertpath_fixture():
    return os.path.join(FIXTURE_DIR, INVERTED_FIXTURE)


@pytest.fixture
def gradientpath_fixture():
    return os.path.join(FIXTURE_DIR, GRADIENT_FIXTURE)


@pytest.fixture
def projectionpath_fixture():
    return (os.path.join(FIXTURE_DIR, projection_fixture_filename) for projection_fixture_filename in
            PROJECTION_FIXTURE)


@pytest.fixture
def registeredpath_fixture():
    return os.path.join(FIXTURE_DIR, REGISTERED_FIXTURE)


@pytest.fixture
def contrastlimits_fixture():
    return PIX_PERCENTILE_LIMITS


@pytest.fixture
def gaussiansigma_fixture():
    return GAUSSIAN_SIGMA


@pytest.fixture
def kymographpath_fixture():
    return os.path.join(FIXTURE_DIR, KYMOGRAPH_FIXTURE)


@pytest.fixture
def rescaletuple_fixture():
    return RESCALE_TUPLE


@pytest.fixture
def zprojectionspath_fixture():
    return os.path.join(FIXTURE_DIR, ZPROJECTIONS_FIXTURE[0]), os.path.join(FIXTURE_DIR, ZPROJECTIONS_FIXTURE[1])


@pytest.fixture
def findseeds_parameters():
    return 1, 1, FINDSEEDS_SIGMA, FINDSEEDS_WINDOWSIZE, FINDSEEDS_BINCLOSINGS, FINDSEEDS_MINDIST


@pytest.fixture
def findseedsannotationspath_fixture():
    return os.path.join(FIXTURE_DIR, FINDSEEDS_ANNOTATIONSPATH)


@pytest.fixture
def propagateseeds_parameters():
    return 1, 20, FINDSEEDS_WINDOWSIZE


@pytest.fixture
def propagateseeds0annotationspath_fixture():
    return os.path.join(FIXTURE_DIR, PROPAGATESEEDS0_ANNOTATIONSPATH)


@pytest.fixture
def propagateseedsannotationspath_fixture():
    return os.path.join(FIXTURE_DIR, PROPAGATESEEDS_ANNOTATIONSPATH)


@pytest.fixture
def expandseedsannotationspath_fixture():
    return os.path.join(FIXTURE_DIR, EXPANDSEEDS_ANNOTATIONSPATH)


@pytest.fixture
def expandseeds_parameters():
    return 1, 1, FINDSEEDS_SIGMA


@pytest.fixture
def expandnpropagateseedsannotationspath_fixture():
    return os.path.join(FIXTURE_DIR, EXPANDNPROPAGATESEEDS_ANNOTATIONSPATH)


@pytest.fixture
def expandnpropagateseeds_parameters():
    return 1, 20, FINDSEEDS_SIGMA, FINDSEEDS_WINDOWSIZE


@pytest.fixture
def projcatfolder_fixture():
    return os.path.join(FIXTURE_DIR, PROJECT_CONCAT_DIR)


@pytest.fixture
def batchmeasurefolder_fixture():
    return os.path.join(FIXTURE_DIR, BATCH_MEASURE_DIR)

@pytest.fixture
def fieldcorrection_fixture():
    return {'input_folder': os.path.join(FIXTURE_DIR, BATCH_FLATFIELD_DIR),
            'darkfield_file': os.path.join(FIXTURE_DIR, BATCH_FLATFIELD_DARKIMAGEPATH),
            'flatfield_file': os.path.join(FIXTURE_DIR, BATCH_FLATFIELD_FLATFIELDIMAGEPATH),
            'crop_dims': BATCH_FLATFIELD_CROPDIMS,
            'file_suffix': BATCH_FLATFIELD_FILESUFFIX,
            'input_substr': BATCH_FLATFIELD_SUBSTR,
            'output_folder': os.path.join(FIXTURE_DIR, BATCH_FLATFIELD_OUTPUTDIR)
            }




"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from abc import ABC, abstractmethod
from typing import Optional, Tuple

from PyQt6 import QtWidgets

from pyjamas.rimage.rimml.rimclassifier import rimclassifier
import pyjamas.rutils as rutils


class ClassifierDialogABC(ABC):
    positive_training_folder: str = ''
    negative_training_folder: str = ''
    hard_negative_training_folder: str = ''
    train_image_size: Tuple[int, int] = (0, 0)
    step_sz: Tuple[int, int] = rimclassifier.DEFAULT_STEP_SZ
    histogram_of_gradients: bool = True

    def __init__(self):
        super().__init__()

        self.positive_training_folder_edit: QtWidgets.QLineEdit = None
        self.negative_training_folder_edit: QtWidgets.QLineEdit = None
        self.hard_negative_training_folder_edit: QtWidgets.QLineEdit = None

    @abstractmethod
    def setupUi(self, NNMLP: QtWidgets.QDialog, parameters: Optional[dict] = None):
        pass

    @abstractmethod
    def retranslateUi(self, NNMLP: QtWidgets.QDialog):
        pass

    @abstractmethod
    def parameters(self) -> dict:
        pass

    def _open_positive_folder_dialog(self) -> bool:
        folder = rutils.RUtils.open_folder_dialog("Positive training folder", ClassifierDialogABC.positive_training_folder)

        if folder == '' or folder is False or self.positive_training_folder_edit is None:
            return False

        self.positive_training_folder_edit.setText(folder)

        return True

    def _open_negative_folder_dialog(self) -> bool:
        folder = rutils.RUtils.open_folder_dialog("Negative training folder", ClassifierDialogABC.negative_training_folder)

        if folder == '' or folder is False or self.negative_training_folder_edit is None:
            return False

        self.negative_training_folder_edit.setText(folder)

        return True

    def _open_hard_folder_dialog(self) -> bool:
        folder = rutils.RUtils.open_folder_dialog("Hard negative training folder", ClassifierDialogABC.hard_negative_training_folder)

        if folder == '' or folder is False or self.hard_negative_training_folder_edit is None:
            return False

        self.hard_negative_training_folder_edit.setText(folder)

        return True




"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
from typing import List, Optional, Tuple

from PyQt6 import QtCore, QtWidgets

from pyjamas.rutils import RUtils


class BatchFlatFieldCorrectionDialog(object):
    BG_MODES: List[str] = ['none', 'mode']
    MAX_DIM: int = 1200

    input_folder: str = None
    darkfield_file: str = None
    flatfield_file: str = None
    crop_dims: Tuple[int, int] = (MAX_DIM//2, MAX_DIM//2)
    input_substr: str = ''
    file_suffix: str = None
    bg_mode: str = BG_MODES[-1]

    def setupUi(self, Dialog, parameters: Optional[dict] = None):
        from pyjamas.rcallbacks.rcbbatchprocess import RCBBatchProcess

        if (parameters is None or parameters is False) and BatchFlatFieldCorrectionDialog.input_folder is None:
            parameters = RCBBatchProcess._default_batchflatfield_parameters()

        if parameters is not None and parameters is not False:
            BatchFlatFieldCorrectionDialog.input_folder = parameters.get('input_folder')
            BatchFlatFieldCorrectionDialog.darkfield_file = parameters.get('darkfield_file')
            BatchFlatFieldCorrectionDialog.flatfield_file = parameters.get('flatfield_file')
            BatchFlatFieldCorrectionDialog.crop_dims = parameters.get('crop_dims')
            BatchFlatFieldCorrectionDialog.bg_mode = parameters.get('bg_mode')
            BatchFlatFieldCorrectionDialog.input_substr = parameters.get('input_substr')
            BatchFlatFieldCorrectionDialog.file_suffix = parameters.get('file_suffix')

        Dialog.setObjectName("Dialog")
        Dialog.resize(609, 303)
        self.ok_buttonbox = QtWidgets.QDialogButtonBox(parent=Dialog)
        self.ok_buttonbox.setGeometry(QtCore.QRect(230, 250, 341, 32))
        self.ok_buttonbox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.ok_buttonbox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.ok_buttonbox.setObjectName("ok_buttonbox")
        self.input_lineedit = QtWidgets.QLineEdit(parent=Dialog)
        self.input_lineedit.setGeometry(QtCore.QRect(160, 30, 341, 21))
        self.input_lineedit.setObjectName("input_lineedit")
        self.input_button = QtWidgets.QPushButton(parent=Dialog)
        self.input_button.setGeometry(QtCore.QRect(510, 25, 61, 32))
        self.input_button.setObjectName("input_button")
        self.input_button.clicked.connect(self._open_input_folder_dialog)
        self.input_label = QtWidgets.QLabel(parent=Dialog)
        self.input_label.setGeometry(QtCore.QRect(40, 33, 91, 16))
        self.input_label.setObjectName("input_label")
        self.darfield_label = QtWidgets.QLabel(parent=Dialog)
        self.darfield_label.setGeometry(QtCore.QRect(40, 68, 101, 16))
        self.darfield_label.setObjectName("darfield_label")
        self.darkfield_lineedit = QtWidgets.QLineEdit(parent=Dialog)
        self.darkfield_lineedit.setGeometry(QtCore.QRect(160, 65, 341, 21))
        self.darkfield_lineedit.setObjectName("darkfield_lineedit")
        self.darkfield_button = QtWidgets.QPushButton(parent=Dialog)
        self.darkfield_button.setGeometry(QtCore.QRect(510, 60, 61, 32))
        self.darkfield_button.setObjectName("darkfield_button")
        self.darkfield_button.clicked.connect(self._open_darkfile_dialog)
        self.flatfield_lineedit = QtWidgets.QLineEdit(parent=Dialog)
        self.flatfield_lineedit.setGeometry(QtCore.QRect(160, 97, 341, 21))
        self.flatfield_lineedit.setObjectName("flatfield_lineedit")
        self.flatfield_button = QtWidgets.QPushButton(parent=Dialog)
        self.flatfield_button.setGeometry(QtCore.QRect(510, 92, 61, 32))
        self.flatfield_button.setObjectName("flatfield_button")
        self.flatfield_button.clicked.connect(self._open_flatfile_dialog)
        self.flatfield_label = QtWidgets.QLabel(parent=Dialog)
        self.flatfield_label.setGeometry(QtCore.QRect(40, 100, 101, 16))
        self.flatfield_label.setObjectName("flatfield_label")
        self.cropdim_label = QtWidgets.QLabel(parent=Dialog)
        self.cropdim_label.setGeometry(QtCore.QRect(40, 130, 101, 16))
        self.cropdim_label.setObjectName("cropdim_label")
        self.rowdim_spinbox = QtWidgets.QSpinBox(parent=Dialog)
        self.rowdim_spinbox.setGeometry(QtCore.QRect(210, 128, 104, 26))
        self.rowdim_spinbox.setObjectName("rowdim_spinbox")
        self.rowdim_spinbox.setRange(0, BatchFlatFieldCorrectionDialog.MAX_DIM)
        self.coldim_spinbox = QtWidgets.QSpinBox(parent=Dialog)
        self.coldim_spinbox.setGeometry(QtCore.QRect(391, 126, 104, 26))
        self.coldim_spinbox.setObjectName("coldim_spinbox")
        self.coldim_spinbox.setRange(0, BatchFlatFieldCorrectionDialog.MAX_DIM)
        self.rowdim_label = QtWidgets.QLabel(parent=Dialog)
        self.rowdim_label.setGeometry(QtCore.QRect(170, 130, 50, 16))
        self.rowdim_label.setObjectName("rowdim_label")
        self.coldim_label = QtWidgets.QLabel(parent=Dialog)
        self.coldim_label.setGeometry(QtCore.QRect(330, 130, 50, 16))
        self.coldim_label.setObjectName("coldim_label")
        self.background_label = QtWidgets.QLabel(parent=Dialog)
        self.background_label.setGeometry(QtCore.QRect(40, 163, 101, 16))
        self.background_label.setObjectName("background_label")
        self.background_combo = QtWidgets.QComboBox(parent=Dialog)
        self.background_combo.setGeometry(QtCore.QRect(160, 159, 131, 26))
        self.background_combo.setObjectName("background_combo")
        for amode in BatchFlatFieldCorrectionDialog.BG_MODES:
            self.background_combo.addItem(amode)
        self.inputsubstr_label = QtWidgets.QLabel(parent=Dialog)
        self.inputsubstr_label.setGeometry(QtCore.QRect(40, 203, 101, 16))
        self.inputsubstr_label.setObjectName("inputsubstr_label")
        self.inputsubstr_lineedit = QtWidgets.QLineEdit(parent=Dialog)
        self.inputsubstr_lineedit.setGeometry(QtCore.QRect(160, 200, 120, 21))
        self.inputsubstr_lineedit.setText("")
        self.inputsubstr_lineedit.setObjectName("inputsubstr_lineedit")
        self.suffix_label = QtWidgets.QLabel(parent=Dialog)
        self.suffix_label.setGeometry(QtCore.QRect(320, 203, 101, 16))
        self.suffix_label.setObjectName("suffix_label")
        self.suffix_lineedit = QtWidgets.QLineEdit(parent=Dialog)
        self.suffix_lineedit.setGeometry(QtCore.QRect(440, 200, 120, 21))
        self.suffix_lineedit.setText("")
        self.suffix_lineedit.setObjectName("suffix_lineedit")

        self.retranslateUi(Dialog)
        self.ok_buttonbox.accepted.connect(Dialog.accept) # type: ignore
        self.ok_buttonbox.rejected.connect(Dialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Batch Flat Field Correction"))
        self.input_button.setText(_translate("Dialog", "..."))
        self.input_label.setText(_translate("Dialog", "input folder"))
        self.darfield_label.setText(_translate("Dialog", "darkfield image"))
        self.darkfield_button.setText(_translate("Dialog", "..."))
        self.flatfield_button.setText(_translate("Dialog", "..."))
        self.flatfield_label.setText(_translate("Dialog", "flatfield image"))
        self.cropdim_label.setText(_translate("Dialog", "crop dimensions"))
        self.rowdim_label.setText(_translate("Dialog", "rows"))
        self.coldim_label.setText(_translate("Dialog", "columns"))
        self.background_label.setText(_translate("Dialog", "background"))
        self.inputsubstr_label.setText(_translate("Dialog", "input substring"))
        self.suffix_label.setText(_translate("Dialog", "output file suffix"))

        self.input_lineedit.setText(BatchFlatFieldCorrectionDialog.input_folder)
        self.darkfield_lineedit.setText(BatchFlatFieldCorrectionDialog.darkfield_file)
        self.flatfield_lineedit.setText(BatchFlatFieldCorrectionDialog.flatfield_file)
        self.rowdim_spinbox.setValue(BatchFlatFieldCorrectionDialog.crop_dims[0])
        self.coldim_spinbox.setValue(BatchFlatFieldCorrectionDialog.crop_dims[1])
        self.background_combo.setCurrentIndex(BatchFlatFieldCorrectionDialog.BG_MODES.index(BatchFlatFieldCorrectionDialog.bg_mode))
        self.inputsubstr_lineedit.setText(BatchFlatFieldCorrectionDialog.input_substr)
        self.suffix_lineedit.setText(BatchFlatFieldCorrectionDialog.file_suffix)

    def parameters(self) -> dict:
        BatchFlatFieldCorrectionDialog.input_folder = self.input_lineedit.text()
        BatchFlatFieldCorrectionDialog.darkfield_file = self.darkfield_lineedit.text()
        BatchFlatFieldCorrectionDialog.flatfield_file = self.flatfield_lineedit.text()
        BatchFlatFieldCorrectionDialog.crop_dims = (self.rowdim_spinbox.value(), self.coldim_spinbox.value())
        BatchFlatFieldCorrectionDialog.bg_mode = self.BG_MODES[self.background_combo.currentIndex()]
        BatchFlatFieldCorrectionDialog.file_suffix = self.suffix_lineedit.text()
        BatchFlatFieldCorrectionDialog.input_substr = self.inputsubstr_lineedit.text()

        return {
            'input_folder': BatchFlatFieldCorrectionDialog.input_folder,
            'darkfield_file': BatchFlatFieldCorrectionDialog.darkfield_file,
            'flatfield_file': BatchFlatFieldCorrectionDialog.flatfield_file,
            'crop_dims': BatchFlatFieldCorrectionDialog.crop_dims,
            'bg_mode': BatchFlatFieldCorrectionDialog.bg_mode,
            'input_substr': BatchFlatFieldCorrectionDialog.input_substr,
            'file_suffix': BatchFlatFieldCorrectionDialog.file_suffix,
        }

    def _open_input_folder_dialog(self) -> bool:
        start_folder = self.input_lineedit.text() if self.input_lineedit.text() != '' else BatchFlatFieldCorrectionDialog.input_folder
        folder = RUtils.open_folder_dialog(f"Input folder", start_folder)

        if folder == '' or folder is False:
            return False

        self.input_lineedit.setText(folder)

        return True

    def _open_darkfile_dialog(self) -> bool:
        start_folder, _ = os.path.split(self.darkfield_lineedit.text())
        if start_folder == '':
            start_folder = os.path.split(BatchFlatFieldCorrectionDialog.darkfield_file)
        fname = QtWidgets.QFileDialog.getOpenFileName(None, 'Dark field file ...',
                                                      BatchFlatFieldCorrectionDialog.input_folder,
                                                      filter='TIFF files (*.tif *.tiff)')

        self.darkfield_lineedit.setText(fname[0])

        return True

    def _open_flatfile_dialog(self) -> bool:
        start_folder, _ = os.path.split(self.flatfield_lineedit.text())
        if start_folder == '':
            start_folder = os.path.split(BatchFlatFieldCorrectionDialog.flatfield_file)
        fname = QtWidgets.QFileDialog.getOpenFileName(None, 'Flat field file ...',
                                                      BatchFlatFieldCorrectionDialog.input_folder,
                                                      filter='TIFF files (*.tif *.tiff)')

        self.flatfield_lineedit.setText(fname[0])

        return True



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
from typing import Optional, Tuple

import matplotlib.pyplot as plt
import skimage.io as sio
import skimage.morphology as sm
import skimage.segmentation as ss
import skimage.transform as st
import tensorflow as tf
import tensorflow.keras.backend as kb
import tensorflow.keras.layers as kl
import tensorflow.keras.models as km
import tensorflow.keras.optimizers as ko
import tensorflow.keras.utils as ku
import numpy

import pyjamas.pjscore
from pyjamas.rimage.rimutils import rimutils
from pyjamas.rimage.rimml.rimneuralnet import rimneuralnet
from pyjamas.rutils import RUtils
from pyjamas.rimage.rimml.classifier_types import classifier_types


class UNet(rimneuralnet):

    EROSION_WIDTH: int = 0
    CLASSIFIER_TYPE: str = classifier_types.UNET.value
    VALIDATION_SPLIT: float = 0.1

    def __init__(self, parameters: Optional[dict] = None):
        super().__init__(parameters)

        self.W_train: numpy.ndarray = None

        output_classes: int = parameters.get('output_classes', UNet.OUTPUT_CLASSES)
        learning_rate: float = parameters.get('learning_rate', UNet.LEARNING_RATE)

        input_size: Tuple[int, int, int] = parameters.get('train_image_size', UNet.TRAIN_IMAGE_SIZE)

        self.epochs: int = parameters.get('epochs', UNet.EPOCHS)
        self.mini_batch_size: int = parameters.get('mini_batch_size', UNet.BATCH_SIZE)
        self.EROSION_WIDTH: int = parameters.get('erosion_width', self.EROSION_WIDTH)

        classifier_representation = parameters.get('classifier')
        if type(classifier_representation) is km.Model:
            self.classifier = classifier_representation
        else:
            if len(input_size) == 2:
                input_size = input_size + (1, )

            self.classifier = self.make_weighted_loss_unet(input_size, output_classes)

        adam = ko.Adam(learning_rate=learning_rate)
        self.classifier.compile(adam, loss=self.pixelwise_loss)

        if type(classifier_representation) is list:
            self.classifier.set_weights(classifier_representation)

        self.step_sz = parameters.get('step_sz', UNet.STEP_SIZE)

    def make_weighted_loss_unet(self, input_shape: Tuple, n_classes: int) -> km.Model:
        _epsilon = tf.convert_to_tensor(kb.epsilon(), numpy.float32)

        # two inputs, one for the image and one for the weight maps
        ip = tf.keras.Input(shape=input_shape, name="image_input")
        # the shape of the weight maps has to be such that it can be element-wise
        # multiplied to the softmax output.
        weight_ip = tf.keras.Input(shape=input_shape[:2] + (n_classes,))

        # adding the layers
        conv1 = kl.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(ip)
        conv1 = kl.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv1)
        conv1 = kl.Dropout(0.1)(conv1)
        mpool1 = kl.MaxPool2D()(conv1)

        conv2 = kl.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(mpool1)
        conv2 = kl.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv2)
        conv2 = kl.Dropout(0.2)(conv2)
        mpool2 = kl.MaxPool2D()(conv2)

        conv3 = kl.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(mpool2)
        conv3 = kl.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv3)
        conv3 = kl.Dropout(0.3)(conv3)
        mpool3 = kl.MaxPool2D()(conv3)

        conv4 = kl.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(mpool3)
        conv4 = kl.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv4)
        conv4 = kl.Dropout(0.4)(conv4)
        mpool4 = kl.MaxPool2D()(conv4)

        conv5 = kl.Conv2D(1024, 3, activation='relu', padding='same', kernel_initializer='he_normal')(mpool4)
        conv5 = kl.Conv2D(1024, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv5)
        conv5 = kl.Dropout(0.5)(conv5)

        up6 = kl.Conv2DTranspose(512, 2, strides=2, kernel_initializer='he_normal', padding='same')(conv5)
        conv6 = kl.Concatenate()([up6, conv4])
        conv6 = kl.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv6)
        conv6 = kl.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv6)
        conv6 = kl.Dropout(0.4)(conv6)

        up7 = kl.Conv2DTranspose(256, 2, strides=2, kernel_initializer='he_normal', padding='same')(conv6)
        conv7 = kl.Concatenate()([up7, conv3])
        conv7 = kl.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv7)
        conv7 = kl.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv7)
        conv7 = kl.Dropout(0.3)(conv7)

        up8 = kl.Conv2DTranspose(128, 2, strides=2, kernel_initializer='he_normal', padding='same')(conv7)
        conv8 = kl.Concatenate()([up8, conv2])
        conv8 = kl.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv8)
        conv8 = kl.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv8)
        conv8 = kl.Dropout(0.2)(conv8)

        up9 = kl.Conv2DTranspose(64, 2, strides=2, kernel_initializer='he_normal', padding='same')(conv8)
        conv9 = kl.Concatenate()([up9, conv1])
        conv9 = kl.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv9)
        conv9 = kl.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv9)
        conv9 = kl.Dropout(0.1)(conv9)

        c10 = kl.Conv2D(n_classes, 1, activation='softmax', kernel_initializer='he_normal', name="unet-activation")(
            conv9)

        # Add a few non trainable layers to mimic the computation of the crossentropy
        # loss, so that the actual loss function just has to perform the
        # aggregation.
        c11 = kl.Lambda(lambda x: x / tf.reduce_sum(x, len(x.get_shape()) - 1, True))(c10)
        c11 = kl.Lambda(lambda x: tf.clip_by_value(x, _epsilon, 1. - _epsilon))(c11)
        c11 = kl.Lambda(lambda x: kb.log(x))(c11)
        weighted_sm = kl.multiply([c11, weight_ip])

        return km.Model(inputs=[ip, weight_ip], outputs=[weighted_sm])

    def pixelwise_loss(self, target, output):
        """
        A custom function defined to simply sum the pixelwise loss.
        This function doesn't compute the crossentropy loss, since that is made a
        part of the model's computational graph itself.
        Parameters
        ----------
        target : tf.tensor
            A tensor corresponding to the true labels of an image.
        output : tf.tensor
            Model output
        Returns
        -------
        tf.tensor
            A tensor holding the aggregated loss.
        """
        return - tf.reduce_sum(target * output,
                               len(output.get_shape()) - 1)

    def fit(self) -> bool:
        self._process_training_data()

        ycat = ku.to_categorical(self.Y_train)
        wmap = numpy.zeros((self.X_train.shape[0], self.train_image_size[0], self.train_image_size[1], 2),
                           dtype=numpy.float32)
        wmap[..., 0] = self.W_train.squeeze()
        wmap[..., 1] = self.W_train.squeeze()

        self.classifier.fit([self.X_train, wmap], ycat, batch_size=self.mini_batch_size, epochs=self.epochs,
                            validation_split=self.VALIDATION_SPLIT)

        return True

    @classmethod
    def weight_map(cls, binmasks: numpy.ndarray, w0: float = 10., sigma: float = 5., show: bool = False):
        """Compute the weight map for a given mask, as described in Ronneberger et al.
        (https://arxiv.org/pdf/1505.04597.pdf)
        """

        labmasks = sm.label(binmasks)
        n_objs = numpy.amax(labmasks)

        nrows, ncols = labmasks.shape[:2]
        masks = numpy.zeros((n_objs, nrows, ncols))
        distMap = numpy.zeros((nrows * ncols, n_objs))
        X1, Y1 = numpy.meshgrid(numpy.arange(nrows), numpy.arange(ncols))
        X1, Y1 = numpy.c_[X1.ravel(), Y1.ravel()].T
        for i in range(n_objs):
            mask = numpy.squeeze(labmasks == i + 1)
            bounds = ss.find_boundaries(mask, mode='inner')
            X2, Y2 = numpy.nonzero(bounds)
            xSum = (X2.reshape(-1, 1) - X1.reshape(1, -1)) ** 2
            ySum = (Y2.reshape(-1, 1) - Y1.reshape(1, -1)) ** 2
            distMap[:, i] = numpy.sqrt(xSum + ySum).min(axis=0)
            masks[i] = mask
        ix = numpy.arange(distMap.shape[0])
        if distMap.shape[1] == 1:
            d1 = distMap.ravel()
            border_loss_map = w0 * numpy.exp((-1 * (d1) ** 2) / (2 * (sigma ** 2)))
        else:
            if distMap.shape[1] == 2:
                d1_ix, d2_ix = numpy.argpartition(distMap, 1, axis=1)[:, :2].T
            else:
                d1_ix, d2_ix = numpy.argpartition(distMap, 2, axis=1)[:, :2].T
            d1 = distMap[ix, d1_ix]
            d2 = distMap[ix, d2_ix]
            border_loss_map = w0 * numpy.exp((-1 * (d1 + d2) ** 2) / (2 * (sigma ** 2)))
        xBLoss = numpy.zeros((nrows, ncols))
        xBLoss[X1, Y1] = border_loss_map
        # class weight map
        loss = numpy.zeros((nrows, ncols))
        w_1 = 1 - masks.sum() / loss.size
        w_0 = 1 - w_1
        loss[masks.sum(0) == 1] = w_1
        loss[masks.sum(0) == 0] = w_0
        ZZ = xBLoss + loss
        # ZZ = resize(ZZ, outsize, preserve_range=True)
        if show:
            plt.imshow(ZZ)
            plt.colorbar()
            plt.axis('off')
        return ZZ

    def _process_training_data(self) -> bool:
        """
        Loads and scales training data and calculates weight maps.
        :return:
        """
        train_ids = next(os.walk(self.positive_training_folder))[1]

        # Get and resize train images and masks
        self.X_train = numpy.zeros((len(train_ids), self.train_image_size[0], self.train_image_size[1], 1), dtype=numpy.uint16)
        self.Y_train = numpy.zeros((len(train_ids), self.train_image_size[0], self.train_image_size[1], 1), dtype=bool)
        self.W_train = numpy.zeros((len(train_ids), self.train_image_size[0], self.train_image_size[1], 1), dtype=float)
        print('Getting and resizing train images and masks ... ')

        for n, id_ in enumerate(train_ids):
            path = self.positive_training_folder + id_
            im_file = path + "/image/" + os.listdir(path + "/image/")[0]
            img = sio.imread(im_file)
            if img.ndim == 3:
                img = img[0, :, :]
            img = numpy.expand_dims(st.resize(img, (self.train_image_size[0], self.train_image_size[1]), order=3,
                                              mode='constant', preserve_range=True), axis=-1)
            self.X_train[n] = img
            msk_file = path + "/mask/" + os.listdir(path + "/mask/")[0]
            mask = numpy.zeros((self.train_image_size[0], self.train_image_size[1], 1), dtype=bool)
            mask_ = sio.imread(msk_file)
            mask_ = numpy.expand_dims(st.resize(mask_, (self.train_image_size[0], self.train_image_size[1]), order=3,
                                                mode='constant', preserve_range=True), axis=-1)
            mask = numpy.maximum(mask, mask_)
            weights = UNet.weight_map(mask)
            self.Y_train[n] = mask
            self.W_train[n, :, :, 0] = weights

        self.scaler = numpy.amax(self.X_train)
        self.X_train = self.X_train / self.scaler

        return True

    def predict(self, image: numpy.ndarray) -> (numpy.ndarray, numpy.ndarray):
        if image is None or image is False:
            return False

        if image.ndim == 3:
            image = image[0, :, :]

        testImage = image / self.scaler

        image_input = self.classifier.get_layer('image_input').input
        softmax_output = self.classifier.get_layer('unet-activation').output
        predictor = kb.function([image_input], [softmax_output])

        testLabel = numpy.zeros(testImage.shape, dtype=bool)
        testProb = numpy.zeros(testImage.shape, dtype=bool)
        half_width = int(self.train_image_size[1] / 2)
        half_height = int(self.train_image_size[0] / 2)

        for animage, therow, thecol in rimutils.generate_subimages(testImage, self.train_image_size[0:2],
                                                                   self.step_sz, True):
            yhat = predictor([numpy.expand_dims(animage, axis=0)])[0]
            yhat = numpy.argmax(yhat[0], axis=-1)
            p = numpy.amax(yhat[0], axis=-1)

            testLabel[(therow - half_height):(therow + half_height),
            (thecol - half_width):(thecol + half_width)] = numpy.logical_or(
                testLabel[(therow - half_height):(therow + half_height), (thecol - half_width):(thecol + half_width)],
                yhat)
            testProb[(therow - half_height):(therow + half_height),
            (thecol - half_width):(thecol + half_width)] = p  # This is not really correct: one should select the probability that makes the pixel get its final value (or an average of those).

        if self.EROSION_WIDTH is not None and self.EROSION_WIDTH != 0:
            self.object_array = numpy.asarray(rimutils.extract_contours(
                sm.dilation(sm.label(sm.binary_erosion(testLabel, sm.square(self.EROSION_WIDTH)), connectivity=1),
                            sm.square(self.EROSION_WIDTH))), dtype=object)
        else:
            self.object_array = numpy.asarray(rimutils.extract_contours(sm.label(testLabel, connectivity=1)),
                                              dtype=object)
        self.prob_array = testProb

        return self.object_array.copy(), self.prob_array.copy()

    def save(self, filename: str) -> bool:
        classifier = self.classifier.get_weights() if self.classifier.weights else self.classifier

        theclassifier = {
            'classifier_type': self.CLASSIFIER_TYPE,
            'positive_training_folder': self.positive_training_folder,
            'train_image_size': self.train_image_size,
            'scaler': self.scaler,
            'epochs': self.epochs,
            'mini_batch_size': self.mini_batch_size,
            'learning_rate': kb.eval(self.classifier.optimizer.lr),
            'classifier': self.classifier.get_weights(),
            'step_sz': self.step_sz,
            'erosion_width': self.EROSION_WIDTH
        }

        return RUtils.pickle_this(theclassifier, RUtils.set_extension(filename, pyjamas.pjscore.PyJAMAS.classifier_extension))



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
from typing import Callable, Optional

import matplotlib.pyplot as plt
import matplotlib as mpl
from PyQt6 import QtWidgets

from pyjamas.pjscore import undo_modes
from pyjamas.rcallbacks.rcallback import RCallback
from pyjamas.rimage.rimcore import rimage

class RCBOptions(RCallback):
    def cbUndo(self) -> bool:
        """
        Undo the most recent action: individual fiducials added or deleted; polylines added, deleted or moved; or multiple fiducials or polylines added or deleted simultaneously (e.g. after finding seeds, segmenting cells, or deleting all annotations). Also, image projections, inversion, smoothing, gradient, registration, crop or kymograph.

        :return: True if undo finished correctly, False if there was an error.
        """

        undoInfo = self.pjs.undo_stack.pop()
        if undoInfo is None:
            return False

        changetype = undoInfo.get("changetype")
        frame = undoInfo.get("frame")
        index = undoInfo.get("index")
        details = undoInfo.get("details")

        try:
            if changetype == undo_modes.FIDUCIAL_ADDED:
                self.pjs.fiducials[frame].pop(index)
            elif changetype == undo_modes.FIDUCIAL_DELETED:
                self.pjs.fiducials[frame].insert(index, details)
            elif changetype == undo_modes.POLYLINE_ADDED:
                self.pjs.polylines[frame].pop(index)
                self.pjs.polyline_ids[frame].pop(index)
            elif changetype == undo_modes.POLYLINE_DELETED:
                self.pjs.polylines[frame].append(details)
                self.pjs.polyline_ids[frame].append(index)
            elif changetype == undo_modes.POLYLINE_REPLACED:
                ind_list = [anind for anind in range(len(self.pjs.polyline_ids[frame]))
                            if self.pjs.polyline_ids[frame][anind] == index]
                self.pjs.polylines[frame][ind_list[0]] = details
            elif changetype == undo_modes.MULTI:
                if index[0]:
                    self.pjs.fiducials[frame[0]:frame[1]] = details[0]
                if index[1]:
                    self.pjs.polylines[frame[0]:frame[1]] = details[1]
                if index[2]:
                    self.pjs.polyline_ids[frame[0]:frame[1]] = details[2]
            elif changetype == undo_modes.IMAGE:
                self.pjs.io.cbLoadArray(frame)
                self.pjs.fiducials = details[0]
                self.pjs.polylines = details[1]
                self.pjs.polyline_ids = details[2]
                self.pjs.image.cbGoTo(index)

            if changetype == undo_modes.IMAGE or changetype == undo_modes.MULTI or frame == self.pjs.curslice:
                self.pjs.repaint()

        except Exception as an_exception:
            print(an_exception)
            self.pjs.undo_stack.clear()
            self.pjs.statusbar.showMessage("Something went wrong with the undo function. Undo stack cleared.")

            return False

        return True

    def cbSetBrushSize(self, sz: int=None) -> bool:
        """
        Set the size of the brush used to paint polygons.

        :param sz: brush size; if the value is None a dialog appears.
        :return: True if the brush size was changed, False otherwise.
        """

        brush_size: int = 0
        ok_flag: bool = None

        if sz not in [None, False]:
            brush_size = sz
            ok_flag = True
        else:
            # Read user input for brush size.
            brush_size, ok_flag = QtWidgets.QInputDialog.getInt(None, 'Set brush size: ', 'Enter new size: ',
                    self.pjs.brush_size, 1)


        if ok_flag and brush_size > 0:
            self.pjs.brush_size = brush_size
            self.pjs.repaint()

            return True

        else:
            return False

    def cbDisplayFiducialIDs(self) -> bool:
        """
        Toggle fiducial ids on/off.

        :return: True.
        """
        if self.pjs.display_fiducial_ids:
            self.pjs.display_fiducial_ids = False
        else:
            self.pjs.display_fiducial_ids = True

        self.pjs.repaint()

        return True

    def cbFramesPerSec(self, fps: Optional[int] = None) -> bool:
        """
        Set the number of frames per second to use when playing through the slices.

        :param fps: number of frames per second; if the value is None, a dialog appears.
        :return: True if the number of frames per second was changed, False otherwise.
        """

        thefps: int = 0
        ok_flag: bool = None

        if fps is not None and fps is not False:
            thefps = fps
            ok_flag = True
        else:
            # Read user input for fps.
            thefps, ok_flag = QtWidgets.QInputDialog.getInt(None, 'Set frames per second: ',
                                                            'Enter frames per second: ',
                                                            self.pjs.fps, 1)

        if ok_flag and thefps > 0:
            self.pjs.fps = thefps

            return True

        else:
            return False

    def cbSetCWD(self, folder_name: str = '') -> bool:
        """
        Set the current working directory.

        :param folder_name: absolute path to the current working directory; if the string is empty ('') or None, a dialog appears.
        :return: True if the directory was set (the selected directory must exist), False otherwise.
        """
        if folder_name == '' or folder_name is False or folder_name is None: # When the menu option is clicked on, for some reason that I do not understand, the function is called with filename = False, which causes a bunch of problems.
            folder_name = QtWidgets.QFileDialog.getExistingDirectory(None, 'Set working folder to ...', self.pjs.cwd)

        # If cancel ...
        if folder_name == '':
            return False

        if os.path.exists(folder_name):
            self.pjs.cwd = os.path.abspath(folder_name)
            self.pjs.statusbar.showMessage(f"Working folder set to {self.pjs.cwd}.")

            return True

        else:
            return False

    def cbSetMarginSize(self, margin_size: Optional[int] = None) -> bool:
        """
        Set the size of margins used for cropping.

        :param margin_size: size of margins in pixels.
        :return: True if margin size was changed, False otherwise.
        """
        themargin: int = 0
        ok_flag: bool = True

        if margin_size is None or not margin_size:
            # Read user input for margin size.
            themargin, ok_flag = QtWidgets.QInputDialog.getInt(None, 'Set margin size: ',
                                                               'Enter new size: ', self.pjs.margin_size, 0,
                                                               min(self.pjs.height, self.pjs.width))
        else:
            themargin = margin_size
            ok_flag = True

        if ok_flag and (themargin >= 0):
            self.pjs.margin_size = themargin
            return True
        else:
            return False

    def cbSetImageProcessingCropSize(self, crop_size: Optional[int] = None) -> bool:
        """
        Set the size of the image crop that will be used for expensive computational operations (e.g. balloons).

        :param crop_size: size of the cropped image.
        :return: True if crop size was changed, False otherwise.
        """
        thesize: int = 0
        ok_flag: bool = True

        if crop_size is None or not crop_size:
            # Read user input for margin size.
            thesize, ok_flag = QtWidgets.QInputDialog.getInt(None, 'Set crop size: ',
                                                             'Enter new size: ', self.pjs.balloon_crop_size * 2, 0,
                                                             min(self.pjs.height, self.pjs.width))
        else:
            thesize = crop_size
            ok_flag = True

        if ok_flag and (thesize >= 0):
            self.pjs.balloon_crop_size = int(thesize / 2)
            return True
        else:
            return False

    def cbCropTracked(self):
        """
        Toggles cropping with tracked polylines on or off.

        :return: True.
        """

        self.pjs.crop_tracked_polyline = not self.pjs.crop_tracked_polyline

        if self.pjs.crop_tracked_polyline:
            self.pjs.statusbar.showMessage('Crop tracked polyline turned ON.')
        else:
            self.pjs.statusbar.showMessage('Crop tracked polyline turned OFF.')

        return True

    def cbChangeDisplayTheme(self) -> bool:
        """
        Toggles display between light theme and dark theme.

        :return: True.
        """
        
        if self.pjs.darktheme is False:
            self.pjs.app.setStyleSheet("""
                QMainWindow{
                    background-color: #242424;
                }
                QMenuBar {
                    background-color: #242424;
                    color: white;
                }
                QMenuBar::item::selected {
                    background-color: #6e6e6e;
                    color: white;
                }
                QMenu {
                    background-color: #242424;
                    color: white;
                }
                QMenu::item::selected{
                    background-color: #6e6e6e;
                    color: white;
                }
                QWidget{
                    background: #242424;
                    color: white;
                }
                QPushButton{
                    background-color: #6e6e6e;
                }
                QPushButton:pressed{
                    background-color: #575757;
                }
                QLabel{
                    background-color: transparent;
                }
            """)
            plt.style.use('dark_background')
            mpl.rc('figure', facecolor = '#242424', edgecolor = 'white')
            mpl.rc('axes', facecolor = '#242424', edgecolor = 'white')
            self.pjs.darktheme = True

        else:
            self.pjs.app.setStyleSheet("")
            plt.style.use('default')
            mpl.rcdefaults()
            self.pjs.darktheme = False
        
        return True

    def cbCloseAllPolylines(self):
        """
        Toggles closing all polylines loaded from file ON/OFF.

        :return: True.
        """

        self.pjs.close_all_polylines = not self.pjs.close_all_polylines

        if self.pjs.close_all_polylines:
            self.pjs.statusbar.showMessage('Close all polylines turned ON.')
        else:
            self.pjs.statusbar.showMessage('Close all polylines turned OFF.')

        return True

    def cbSetLiveWireSmoothGaussian(self, sigma: Optional[float] = None) -> bool:
        """
        Set the standard deviation of the Gaussian used to smoothen images for LiveWire segmentation.
        A weight of 0.0 is no smoothing.

        :param sigma: standard deviation of the Gaussian used for image smoothing.
        :return: True if sigma was changed, False otherwise.
        """
        thesigma: float = 0.
        ok_flag: bool = True

        if sigma is None or not sigma:
            # Read user input for margin size.
            thesigma, ok_flag = QtWidgets.QInputDialog.getDouble(None, 'Set Gaussian sigma for LiveWire segmentation: ',
                                                                  'Enter new sigma: ',
                                                                 self.pjs.livewire_gaussian_sigma, 0)
        else:
            thesigma = sigma
            ok_flag = True

        if ok_flag and (thesigma >= 0.):
            self.pjs.livewire_gaussian_sigma = thesigma
            return True
        else:
            return False

    def cbSetLivewireShortestPathFunction(self, shortest_path_fn: Optional[Callable]=None):
        """
        Set the function used to calculate the shortest path between two pixels using LiveWire segmentation.

        :param shortest_path_fn: function to calculate the shortest path between two pixels (see rimage.rimcore.rimage.livewire_shortest_path_fns for examples).
        :return: True if function was changed, False otherwise.
        """
        thefn = None

        if shortest_path_fn is None or not shortest_path_fn:
            fn_names = tuple(rimage.livewire_shortest_path_fns.keys())
            fns = tuple(rimage.livewire_shortest_path_fns.values())
            thefn, ok_flag = QtWidgets.QInputDialog.getItem(None, 'Select shortest path function for LiveWire segmentation: ',
                                            'Function', fn_names, fns.index(self.pjs.livewire_shortest_path_fn), False)
            if thefn is not None:
                thefn = rimage.livewire_shortest_path_fns.get(thefn)
        else:
            thefn = shortest_path_fn
            ok_flag = True

        if ok_flag:
            self.pjs.livewire_shortest_path_fn = thefn
            return True
        else:
            return False


"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from abc import ABC, abstractmethod
from typing import Optional

import numpy

from pyjamas.pjsthreads import ThreadSignals
from pyjamas.rimage.rimml import rimml


class BatchML(ABC):
    def __init__(self, n_frames: int, classifier: rimml = None):
        self.n_frames = n_frames

        if classifier is not None:
            self.image_classifier: rimml = classifier
        else:
            self.image_classifier: rimml = None

    @abstractmethod
    def fit(self, stop_signal: Optional[ThreadSignals] = None) -> bool:
        pass

    @abstractmethod
    def predict(self, slices: numpy.ndarray, indices: numpy.ndarray, progress_signal: Optional[ThreadSignals] = None) -> bool:
        pass

    @abstractmethod
    def reset_objects(self, n_frames: int) -> bool:
        pass



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from functools import partial
from typing import List, Optional

from PyQt6 import QtCore, QtWidgets

from pyjamas.rutils import RUtils


class BatchMeasureDialog(object):
    dirs: List[str] = None
    analyze_flag: bool = None
    analysis_filename_appendix: str = None
    analysis_extension: str = None
    save_results: bool = None  # Save the analysis script in each folder in which an analysis flag is saved.
    script_filename_appendix: str = None
    results_folder: str = None
    intensity_flag: bool = None  # Run intensity section of the analysis/plots?
    image_extension: str = None
    normalize_intensity_flag: int = None
    normalization_modes: List[str] = ['none', 'photobleaching (divide by image mean)',
                                      'image background and photobleaching (subtract image mode and divide by image mean)',
                                      'sample photobleaching (autodetect sample and divide by sample mean)',
                                      'sample background and photobleaching (autodetect sample, subtract sample mode and divide by sample mean)',
                                      'file (load additional image, subtract image mode and divide by image mean)']
    t_res: float = None  # Time resolution in seconds.
    xy_res: float = None  # Spatial resolution in microns.
    index_time_zero: int = None  # Number of time points before treatment (e.g. number of images before wounding).
    max_index_time_zero = 30
    plot_flag: bool = None  # Generate and display plots.
    group_labels: List[str] = None
    err_styles: List[str] = ['band', 'bars']
    err_style_value: str = None
    plot_style_value: str = None
    plot_styles: List[str] = ['box', 'violin']
    brush_sz: int = None
    max_brush_sz: int = 15
    compile_data_flag: bool = None  # Read all data and compile into DataFrames.
    line_width: int = None
    max_line_width: int = 15

    def __init__(self):
        super().__init__()

    def setupUi(self, Dialog, parameters: Optional[dict] = None):
        from pyjamas.rcallbacks.rcbbatchprocess import RCBBatchProcess

        if (parameters is None or parameters is False) and BatchMeasureDialog.dirs is None:
            parameters = RCBBatchProcess._default_batchmeasure_parameters()

        if parameters is not None and parameters is not False:
            BatchMeasureDialog.dirs = parameters.get('folders')
            BatchMeasureDialog.analyze_flag = parameters.get('analyze_flag')
            BatchMeasureDialog.analysis_filename_appendix = parameters.get('analysis_filename_appendix')
            BatchMeasureDialog.analysis_extension = parameters.get('analysis_extension')
            BatchMeasureDialog.save_results = parameters.get('save_results')  # Save the analysis script in each folder in which an analysis flag is saved.
            BatchMeasureDialog.script_filename_appendix = parameters.get('script_filename_appendix')
            BatchMeasureDialog.results_folder = parameters.get('results_folder')
            BatchMeasureDialog.intensity_flag = parameters.get('intensity_flag')  # Run intensity section of the analysis/plots?
            BatchMeasureDialog.image_extension = parameters.get('image_extension')
            BatchMeasureDialog.normalize_intensity_flag = parameters.get('normalize_intensity_flag')
            BatchMeasureDialog.t_res = parameters.get('t_res')  # Time resolution in seconds.
            BatchMeasureDialog.xy_res = parameters.get('xy_res')  # Spatial resolution in microns.
            BatchMeasureDialog.index_time_zero = parameters.get('index_time_zero')  # Number of time points before treatment (e.g. number of images before wounding).
            BatchMeasureDialog.plot_flag = parameters.get('plot_flag')  # Generate and display plots.
            BatchMeasureDialog.group_labels = parameters.get('names')
            BatchMeasureDialog.err_style_value = parameters.get('err_style_value')
            BatchMeasureDialog.plot_style_value = parameters.get('plot_style_value')
            BatchMeasureDialog.brush_sz = parameters.get('brush_sz')

        Dialog.setObjectName("Dialog")
        Dialog.resize(704, 530)
        self.buttonsOkCancel = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonsOkCancel.setGeometry(QtCore.QRect(310, 490, 341, 32))
        self.buttonsOkCancel.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonsOkCancel.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonsOkCancel.setObjectName("buttonsOkCancel")

        self.GroupTabBox = QtWidgets.QTabWidget(Dialog)
        self.GroupTabBox.setGeometry(QtCore.QRect(20, 10, 661, 91))
        self.GroupTabBox.setObjectName("groupTabBox")

        self.dir_groupboxes = []
        self.dir_labels = []
        self.dir_folderlines = []
        self.dir_pbs = []
        self.dir_namelines = []
        self.dir_namelabels = []

        for dir_index, dir_path in enumerate(self.dirs):
            self.dir_groupboxes.append(QtWidgets.QGroupBox())
            self.dir_groupboxes[dir_index].setGeometry(QtCore.QRect(1, 1, 659, 89))
            self.dir_groupboxes[dir_index].setObjectName(f"dir_groupbox_{dir_index}")
            self.dir_labels.append(QtWidgets.QLabel(self.dir_groupboxes[dir_index]))
            self.dir_labels[dir_index].setGeometry(QtCore.QRect(10, 10, 131, 31))
            self.dir_labels[dir_index].setObjectName(f"dir_label_{dir_index}")
            self.dir_folderlines.append(QtWidgets.QLineEdit(self.dir_groupboxes[dir_index]))
            self.dir_folderlines[dir_index].setGeometry(QtCore.QRect(165, 15, 271, 21))
            self.dir_folderlines[dir_index].setObjectName(f"leFolder_{dir_index}")
            self.dir_pbs.append(QtWidgets.QPushButton(self.dir_groupboxes[dir_index]))
            self.dir_pbs[dir_index].setGeometry(QtCore.QRect(432, 11, 51, 32))
            self.dir_pbs[dir_index].setObjectName(f"pbFolder_{dir_index}")
            self.dir_pbs[dir_index].clicked.connect(partial(self._open_folder_dialog, dir_index))
            self.dir_namelines.append(QtWidgets.QLineEdit(self.dir_groupboxes[dir_index]))
            self.dir_namelines[dir_index].setGeometry(QtCore.QRect(550, 15, 101, 21))
            self.dir_namelines[dir_index].setObjectName(f"leName_{dir_index}")
            self.dir_namelabels.append(QtWidgets.QLabel(self.dir_groupboxes[dir_index]))
            self.dir_namelabels[dir_index].setGeometry(QtCore.QRect(499, 10, 131, 31))
            self.dir_namelabels[dir_index].setObjectName(f"name_label_{dir_index}")
            self.GroupTabBox.addTab(self.dir_groupboxes[dir_index], f"{dir_index+1}")
            self.dir_folderlines[dir_index].raise_()
            self.dir_namelines[dir_index].raise_()

        self.addTabPb = QtWidgets.QPushButton(Dialog)
        self.addTabPb.setGeometry(QtCore.QRect(20, 105, 100, 32))
        self.addTabPb.setObjectName("addtab_pb")
        self.addTabPb.clicked.connect(partial(self._add_tab, Dialog))
        self.removeTabPb = QtWidgets.QPushButton(Dialog)
        self.removeTabPb.setGeometry(QtCore.QRect(130, 105, 100, 32))
        self.removeTabPb.setObjectName("removetab_pb")
        self.removeTabPb.clicked.connect(partial(self._remove_tab, Dialog))

        self.groupBox_2 = QtWidgets.QGroupBox(Dialog)
        self.groupBox_2.setGeometry(QtCore.QRect(20, 130, 531, 351))
        self.groupBox_2.setObjectName("groupBox_2")
        self.cbRunAnalysis = QtWidgets.QCheckBox(self.groupBox_2)
        self.cbRunAnalysis.setGeometry(QtCore.QRect(10, 237, 101, 20))
        self.cbRunAnalysis.setObjectName("cbRunAnalysis")
        self.cbAnalyzeIntensities = QtWidgets.QCheckBox(self.groupBox_2)
        self.cbAnalyzeIntensities.setGeometry(QtCore.QRect(30, 84, 181, 20))
        self.cbAnalyzeIntensities.setObjectName("cbAnalyzeIntensities")
        self.label_5 = QtWidgets.QLabel(self.groupBox_2)
        self.label_5.setGeometry(QtCore.QRect(30, 157, 171, 31))
        self.label_5.setObjectName("label_5")
        self.leAnalysisFileSuffix = QtWidgets.QLineEdit(self.groupBox_2)
        self.leAnalysisFileSuffix.setGeometry(QtCore.QRect(210, 162, 271, 21))
        self.leAnalysisFileSuffix.setObjectName("leAnalysisFileSuffix")
        self.leAnalysisFileExtension = QtWidgets.QLineEdit(self.groupBox_2)
        self.leAnalysisFileExtension.setGeometry(QtCore.QRect(210, 187, 271, 21))
        self.leAnalysisFileExtension.setObjectName("leAnalysisFileExtension")
        self.label_6 = QtWidgets.QLabel(self.groupBox_2)
        self.label_6.setGeometry(QtCore.QRect(30, 182, 171, 31))
        self.label_6.setObjectName("label_6")
        self.leScriptFileSuffix = QtWidgets.QLineEdit(self.groupBox_2)
        self.leScriptFileSuffix.setGeometry(QtCore.QRect(200, 302, 271, 21))
        self.leScriptFileSuffix.setObjectName("leScriptFileSuffix")
        self.label_7 = QtWidgets.QLabel(self.groupBox_2)
        self.label_7.setGeometry(QtCore.QRect(50, 272, 171, 31))
        self.label_7.setObjectName("label_7")
        self.cbSaveResults = QtWidgets.QCheckBox(self.groupBox_2)
        self.cbSaveResults.setGeometry(QtCore.QRect(10, 259, 181, 20))
        self.cbSaveResults.setObjectName("cbSaveResults")
        self.label_8 = QtWidgets.QLabel(self.groupBox_2)
        self.label_8.setGeometry(QtCore.QRect(50, 297, 171, 31))
        self.label_8.setObjectName("label_8")
        self.leFolderResults = QtWidgets.QLineEdit(self.groupBox_2)
        self.leFolderResults.setGeometry(QtCore.QRect(200, 277, 271, 21))
        self.leFolderResults.setObjectName("leFolderResults")
        self.pbFolderResults = QtWidgets.QPushButton(Dialog)
        self.pbFolderResults.setGeometry(QtCore.QRect(492, 402, 51, 32))
        self.pbFolderResults.setObjectName("pbFolderResults")
        self.pbFolderResults.clicked.connect(self._open_results_folder_dialog)
        self.leImageExtension = QtWidgets.QLineEdit(self.groupBox_2)
        self.leImageExtension.setGeometry(QtCore.QRect(170, 109, 31, 21))
        self.leImageExtension.setObjectName("leImageExtension")
        self.label_13 = QtWidgets.QLabel(self.groupBox_2)
        self.label_13.setGeometry(QtCore.QRect(60, 104, 171, 31))
        self.label_13.setObjectName("label_13")
        self.label_14 = QtWidgets.QLabel(self.groupBox_2)
        self.label_14.setGeometry(QtCore.QRect(30, 132, 171, 31))
        self.label_14.setObjectName("label_14")
        self.cbNormalization = QtWidgets.QComboBox(self.groupBox_2)
        self.cbNormalization.setGeometry(QtCore.QRect(205, 137, 281, 26))
        self.cbNormalization.setObjectName("cbNormalization")
        for a_mode in BatchMeasureDialog.normalization_modes:
            self.cbNormalization.addItem(a_mode)
        self.label_15 = QtWidgets.QLabel(self.groupBox_2)
        self.label_15.setGeometry(QtCore.QRect(30, 23, 171, 31))
        self.label_15.setObjectName("label_15")
        self.leTimeResolution = QtWidgets.QLineEdit(self.groupBox_2)
        self.leTimeResolution.setGeometry(QtCore.QRect(190, 28, 51, 21))
        self.leTimeResolution.setObjectName("leTimeResolution")
        self.leXYResolution = QtWidgets.QLineEdit(self.groupBox_2)
        self.leXYResolution.setGeometry(QtCore.QRect(190, 57, 51, 21))
        self.leXYResolution.setObjectName("leXYResolution")
        self.label_16 = QtWidgets.QLabel(self.groupBox_2)
        self.label_16.setGeometry(QtCore.QRect(30, 52, 141, 31))
        self.label_16.setObjectName("label_16")
        self.cbPlotResults = QtWidgets.QCheckBox(self.groupBox_2)
        self.cbPlotResults.setGeometry(QtCore.QRect(10, 324, 181, 20))
        self.cbPlotResults.setObjectName("cbPlotResults")
        self.label_17 = QtWidgets.QLabel(self.groupBox_2)
        self.label_17.setGeometry(QtCore.QRect(280, 52, 121, 31))
        self.label_17.setObjectName("label_17")
        self.cbBrushSz = QtWidgets.QComboBox(self.groupBox_2)
        self.cbBrushSz.setGeometry(QtCore.QRect(400, 56, 104, 26))
        self.cbBrushSz.setObjectName("cbBrushSz")
        for i in range(BatchMeasureDialog.max_brush_sz):
            self.cbBrushSz.addItem(str(i+1))
        self.cbErrorStyle = QtWidgets.QComboBox(self.groupBox_2)
        self.cbErrorStyle.setGeometry(QtCore.QRect(120, 212, 104, 26))
        self.cbErrorStyle.setObjectName("cbErrorStyle")
        for a_style in BatchMeasureDialog.err_styles:
            self.cbErrorStyle.addItem(a_style)
        self.label_18 = QtWidgets.QLabel(self.groupBox_2)
        self.label_18.setGeometry(QtCore.QRect(55, 207, 71, 31))
        self.label_18.setObjectName("label_18")
        self.label_40 = QtWidgets.QLabel(self.groupBox_2)
        self.label_40.setGeometry(QtCore.QRect(280, 23, 171, 31))
        self.label_40.setObjectName("label_40")
        self.cbIndexZero = QtWidgets.QComboBox(self.groupBox_2)
        self.cbIndexZero.setGeometry(QtCore.QRect(400, 23, 104, 26))
        self.cbIndexZero.setObjectName("cbIndexZero")
        for i in range(self.max_index_time_zero+1):
            self.cbIndexZero.addItem(str(i))
        self.cbPlotStyle = QtWidgets.QComboBox(self.groupBox_2)
        self.cbPlotStyle.setGeometry(QtCore.QRect(315, 212, 104, 26))
        self.cbPlotStyle.setObjectName("cbPlotStyle")
        for a_style in BatchMeasureDialog.plot_styles:
            self.cbPlotStyle.addItem(a_style)
        self.label_41 = QtWidgets.QLabel(self.groupBox_2)
        self.label_41.setGeometry(QtCore.QRect(250, 207, 71, 31))
        self.label_41.setObjectName("label_41")
        self.label_13.raise_()
        self.cbRunAnalysis.raise_()
        self.cbAnalyzeIntensities.raise_()
        self.label_5.raise_()
        self.leAnalysisFileSuffix.raise_()
        self.leAnalysisFileExtension.raise_()
        self.label_6.raise_()
        self.label_7.raise_()
        self.cbSaveResults.raise_()
        self.label_8.raise_()
        self.leScriptFileSuffix.raise_()
        self.leFolderResults.raise_()
        self.leImageExtension.raise_()
        self.label_14.raise_()
        self.cbNormalization.raise_()
        self.label_15.raise_()
        self.leTimeResolution.raise_()
        self.leXYResolution.raise_()
        self.label_16.raise_()
        self.cbPlotResults.raise_()
        self.label_17.raise_()
        self.cbBrushSz.raise_()
        self.cbErrorStyle.raise_()
        self.label_18.raise_()
        self.label_40.raise_()
        self.cbIndexZero.raise_()
        self.cbPlotStyle.raise_()
        self.label_41.raise_()
        self.groupBox_3 = QtWidgets.QGroupBox(Dialog)
        self.groupBox_3.setGeometry(QtCore.QRect(560, 130, 120, 351))
        self.groupBox_3.setObjectName("groupBox_3")
        self.label_19 = QtWidgets.QLabel(self.groupBox_3)
        self.label_19.setGeometry(QtCore.QRect(10, 20, 51, 31))
        self.label_19.setObjectName("label_19")
        self.label_20 = QtWidgets.QLabel(self.groupBox_3)
        self.label_20.setGeometry(QtCore.QRect(10, 51, 51, 31))
        self.label_20.setObjectName("label_20")
        self.label_21 = QtWidgets.QLabel(self.groupBox_3)
        self.label_21.setGeometry(QtCore.QRect(10, 65, 81, 31))
        self.label_21.setObjectName("label_21")
        self.label_22 = QtWidgets.QLabel(self.groupBox_3)
        self.label_22.setGeometry(QtCore.QRect(10, 81, 91, 31))
        self.label_22.setObjectName("label_22")
        self.label_23 = QtWidgets.QLabel(self.groupBox_3)
        self.label_23.setGeometry(QtCore.QRect(10, 123, 101, 31))
        self.label_23.setObjectName("label_23")
        self.label_24 = QtWidgets.QLabel(self.groupBox_3)
        self.label_24.setGeometry(QtCore.QRect(10, 109, 51, 31))
        self.label_24.setObjectName("label_24")
        self.label_25 = QtWidgets.QLabel(self.groupBox_3)
        self.label_25.setGeometry(QtCore.QRect(10, 139, 101, 31))
        self.label_25.setObjectName("label_25")
        self.label_26 = QtWidgets.QLabel(self.groupBox_3)
        self.label_26.setGeometry(QtCore.QRect(10, 96, 91, 31))
        self.label_26.setObjectName("label_26")
        self.label_27 = QtWidgets.QLabel(self.groupBox_3)
        self.label_27.setGeometry(QtCore.QRect(10, 153, 91, 31))
        self.label_27.setObjectName("label_27")
        self.label_28 = QtWidgets.QLabel(self.groupBox_3)
        self.label_28.setGeometry(QtCore.QRect(10, 166, 51, 31))
        self.label_28.setObjectName("label_28")
        self.label_29 = QtWidgets.QLabel(self.groupBox_3)
        self.label_29.setGeometry(QtCore.QRect(10, 180, 101, 31))
        self.label_29.setObjectName("label_29")
        self.label_30 = QtWidgets.QLabel(self.groupBox_3)
        self.label_30.setGeometry(QtCore.QRect(10, 196, 111, 31))
        self.label_30.setObjectName("label_30")
        self.label_31 = QtWidgets.QLabel(self.groupBox_3)
        self.label_31.setGeometry(QtCore.QRect(10, 212, 91, 31))
        self.label_31.setObjectName("label_31")
        self.label_32 = QtWidgets.QLabel(self.groupBox_3)
        self.label_32.setGeometry(QtCore.QRect(10, 225, 51, 31))
        self.label_32.setObjectName("label_32")
        self.label_33 = QtWidgets.QLabel(self.groupBox_3)
        self.label_33.setGeometry(QtCore.QRect(10, 255, 111, 31))
        self.label_33.setObjectName("label_33")
        self.label_34 = QtWidgets.QLabel(self.groupBox_3)
        self.label_34.setGeometry(QtCore.QRect(10, 239, 111, 31))
        self.label_34.setObjectName("label_34")
        self.label_35 = QtWidgets.QLabel(self.groupBox_3)
        self.label_35.setGeometry(QtCore.QRect(25, 293, 91, 31))
        self.label_35.setObjectName("label_35")
        self.label_36 = QtWidgets.QLabel(self.groupBox_3)
        self.label_36.setGeometry(QtCore.QRect(10, 266, 91, 31))
        self.label_36.setObjectName("label_36")
        self.label_37 = QtWidgets.QLabel(self.groupBox_3)
        self.label_37.setGeometry(QtCore.QRect(25, 309, 91, 31))
        self.label_37.setObjectName("label_37")
        self.label_38 = QtWidgets.QLabel(self.groupBox_3)
        self.label_38.setGeometry(QtCore.QRect(10, 279, 51, 31))
        self.label_38.setObjectName("label_38")
        self.label_39 = QtWidgets.QLabel(self.groupBox_3)
        self.label_39.setGeometry(QtCore.QRect(10, 35, 51, 31))
        self.label_39.setObjectName("label_39")
        self.buttonsOkCancel.raise_()
        self.groupBox_2.raise_()
        self.groupBox_3.raise_()
        self.pbFolderResults.raise_()

        self.retranslateUi(Dialog, overwrite_inputs=True)
        self.buttonsOkCancel.accepted.connect(Dialog.accept)
        self.buttonsOkCancel.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog, overwrite_inputs: bool = True):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Measure"))

        for d, this_dir in enumerate(self.dirs):
            self.dir_labels[d].setText(_translate("Dialog", f"experiment folder {d+1}"))
            self.dir_pbs[d].setText(_translate("Dialog", "..."))
            self.dir_namelabels[d].setText(_translate("Dialog", "name"))
            if overwrite_inputs:
                self.dir_folderlines[d].setText(BatchMeasureDialog.dirs[d])
                self.dir_namelines[d].setText(_translate("Dialog", BatchMeasureDialog.group_labels[d]))

        self.addTabPb.setText(_translate("Dialog", "Add group"))
        self.removeTabPb.setText(_translate("Dialog", "Remove group"))

        self.groupBox_2.setTitle(_translate("Dialog", "Parameters"))
        self.cbRunAnalysis.setText(_translate("Dialog", "run analysis"))
        self.cbAnalyzeIntensities.setText(_translate("Dialog", "analyze image intensities"))
        self.label_5.setText(_translate("Dialog", "analysis file name suffix"))
        self.label_6.setText(_translate("Dialog", "analysis file name extension"))
        self.label_7.setText(_translate("Dialog", "results folder"))
        self.pbFolderResults.setText(_translate("Dialog", "..."))
        self.cbSaveResults.setText(_translate("Dialog", "save analysis results"))
        self.label_8.setText(_translate("Dialog", "notebook suffix"))
        self.label_13.setText(_translate("Dialog", "image extension"))
        self.label_14.setText(_translate("Dialog", "intensity normalization"))
        self.label_15.setText(_translate("Dialog", "time resolution (seconds)"))
        self.label_16.setText(_translate("Dialog", "xy resolution (microns)"))
        self.cbPlotResults.setText(_translate("Dialog", "plot results"))
        self.label_17.setText(_translate("Dialog", "brush size (pixels)"))
        self.label_18.setText(_translate("Dialog", "error style"))
        self.label_40.setText(_translate("Dialog", "images before \ntreatment"))
        self.label_41.setText(_translate("Dialog", "plot style"))
        self.groupBox_3.setTitle(_translate("Dialog", "Sample folder"))
        self.label_19.setText(_translate("Dialog", "control"))
        self.label_20.setText(_translate("Dialog", "|-n1"))
        self.label_21.setText(_translate("Dialog", "|   |-image.tif"))
        self.label_22.setText(_translate("Dialog", "|   |-image.pjs"))
        self.label_23.setText(_translate("Dialog", "|   |-myimage.tif"))
        self.label_24.setText(_translate("Dialog", "|-n2"))
        self.label_25.setText(_translate("Dialog", "|   |-myimage.pjs"))
        self.label_26.setText(_translate("Dialog", "| "))
        self.label_27.setText(_translate("Dialog", "| "))
        self.label_28.setText(_translate("Dialog", "|-n3"))
        self.label_29.setText(_translate("Dialog", "|   |-201914.tif"))
        self.label_30.setText(_translate("Dialog", "|   |-201914.pjs"))
        self.label_31.setText(_translate("Dialog", "| "))
        self.label_32.setText(_translate("Dialog", "|-a3592"))
        self.label_33.setText(_translate("Dialog", "|   |-a3592.pjs"))
        self.label_34.setText(_translate("Dialog", "|   |-a3592.tif"))
        self.label_35.setText(_translate("Dialog", "|-a12_3.tif"))
        self.label_36.setText(_translate("Dialog", "| "))
        self.label_37.setText(_translate("Dialog", "|-a12_3.pjs"))
        self.label_38.setText(_translate("Dialog", "|-exp36"))
        self.label_39.setText(_translate("Dialog", "|"))

        if overwrite_inputs:
            self.cbRunAnalysis.setChecked(BatchMeasureDialog.analyze_flag)
            self.cbAnalyzeIntensities.setChecked(BatchMeasureDialog.intensity_flag)
            self.leAnalysisFileSuffix.setText(_translate("Dialog", BatchMeasureDialog.analysis_filename_appendix))
            self.leAnalysisFileExtension.setText(_translate("Dialog", BatchMeasureDialog.analysis_extension))
            self.leFolderResults.setText(_translate("Dialog", BatchMeasureDialog.results_folder))
            self.cbSaveResults.setChecked(BatchMeasureDialog.save_results)
            self.leScriptFileSuffix.setText(_translate("Dialog", BatchMeasureDialog.script_filename_appendix))
            self.leImageExtension.setText(_translate("Dialog", BatchMeasureDialog.image_extension))
            self.cbNormalization.setCurrentIndex(int(BatchMeasureDialog.normalize_intensity_flag))
            self.leTimeResolution.setText(str(BatchMeasureDialog.t_res))
            self.leXYResolution.setText(str(round(BatchMeasureDialog.xy_res*1000)/1000.))
            self.cbPlotResults.setChecked(BatchMeasureDialog.plot_flag)
            self.cbBrushSz.setCurrentIndex(BatchMeasureDialog.brush_sz-1)
            self.cbErrorStyle.setCurrentIndex(BatchMeasureDialog.err_styles.index(BatchMeasureDialog.err_style_value))
            self.cbIndexZero.setCurrentIndex(BatchMeasureDialog.index_time_zero)
            self.cbPlotStyle.setCurrentIndex(BatchMeasureDialog.plot_styles.index(BatchMeasureDialog.plot_style_value))

    def parameters(self) -> dict:

        BatchMeasureDialog.dirs: List[str] = [this_folder.text() for this_folder in self.dir_folderlines]
        BatchMeasureDialog.analyze_flag: bool = self.cbRunAnalysis.isChecked()
        BatchMeasureDialog.analysis_filename_appendix: str = self.leAnalysisFileSuffix.text()
        BatchMeasureDialog.analysis_extension: str = self.leAnalysisFileExtension.text()
        BatchMeasureDialog.results_folder: str = self.leFolderResults.text()
        BatchMeasureDialog.save_results: bool = self.cbSaveResults.isChecked()  # Save the analysis script in each folder in which an analysis flag is saved.
        BatchMeasureDialog.script_filename_appendix: str = self.leScriptFileSuffix.text()
        BatchMeasureDialog.intensity_flag: bool = self.cbAnalyzeIntensities.isChecked()  # Run intensity section of the analysis/plots?
        BatchMeasureDialog.image_extension: str = self.leImageExtension.text()
        BatchMeasureDialog.normalize_intensity_flag = self.cbNormalization.currentIndex()
        BatchMeasureDialog.t_res: float = float(self.leTimeResolution.text())  # Time resolution in seconds.
        BatchMeasureDialog.xy_res: float = float(self.leXYResolution.text())  # Spatial resolution in microns.
        BatchMeasureDialog.index_time_zero: int = self.cbIndexZero.currentIndex() # Number of time points before treatment (e.g. number of images before wounding).
        BatchMeasureDialog.plot_flag: bool = self.cbPlotResults.isChecked()  # Generate and display plots.
        BatchMeasureDialog.group_labels: List[str] = [this_name.text() for this_name in self.dir_namelines]
        BatchMeasureDialog.err_style_value: str = self.cbErrorStyle.currentText()
        BatchMeasureDialog.plot_style_value: str = self.cbPlotStyle.currentText()
        BatchMeasureDialog.brush_sz: int = int(self.cbBrushSz.currentText())

        return {
            'folders': BatchMeasureDialog.dirs,
            'analyze_flag': BatchMeasureDialog.analyze_flag,
            'analysis_filename_appendix': BatchMeasureDialog.analysis_filename_appendix,
            'analysis_extension': BatchMeasureDialog.analysis_extension,
            'save_results': BatchMeasureDialog.save_results,
            'script_filename_appendix': BatchMeasureDialog.script_filename_appendix,
            'results_folder': BatchMeasureDialog.results_folder,
            'intensity_flag': BatchMeasureDialog.intensity_flag,
            'image_extension': BatchMeasureDialog.image_extension,
            'normalize_intensity_flag': BatchMeasureDialog.normalize_intensity_flag,
            't_res': BatchMeasureDialog.t_res,
            'xy_res': BatchMeasureDialog.xy_res,
            'index_time_zero': BatchMeasureDialog.index_time_zero,
            'plot_flag': BatchMeasureDialog.plot_flag,
            'names': BatchMeasureDialog.group_labels,
            'err_style_value': BatchMeasureDialog.err_style_value,
            'plot_style_value': BatchMeasureDialog.plot_style_value,
            'brush_sz': BatchMeasureDialog.brush_sz,
        }

    def _open_folder_dialog(self, dir_num: int = 0) -> bool:
        start_folder = self.dir_folderlines[dir_num].text() if self.dir_folderlines[dir_num].text() != '' else BatchMeasureDialog.dirs[dir_num]
        folder = RUtils.open_folder_dialog(f"{BatchMeasureDialog.group_labels[dir_num]} folder", start_folder)

        if folder == '' or folder is False:
            return False

        self.dir_folderlines[dir_num].setText(folder)
        return True

    def _add_tab(self, Dialog):
        self.dirs.append("")
        self.group_labels.append(f"group {len(self.dirs)}")
        self.dir_groupboxes.append(QtWidgets.QGroupBox())
        self.dir_groupboxes[-1].setGeometry(QtCore.QRect(1, 1, 659, 89))
        self.dir_groupboxes[-1].setObjectName(f"dir_groupbox_{len(self.dir_groupboxes) - 1}")
        self.dir_labels.append(QtWidgets.QLabel(self.dir_groupboxes[-1]))
        self.dir_labels[-1].setGeometry(QtCore.QRect(10, 10, 131, 31))
        self.dir_labels[-1].setObjectName(f"dir_label_{len(self.dir_labels) - 1}")
        self.dir_folderlines.append(QtWidgets.QLineEdit(self.dir_groupboxes[-1]))
        self.dir_folderlines[-1].setGeometry(QtCore.QRect(165, 15, 271, 21))
        self.dir_folderlines[-1].setObjectName(f"leFolder_{len(self.dir_folderlines) - 1}")
        self.dir_pbs.append(QtWidgets.QPushButton(self.dir_groupboxes[-1]))
        self.dir_pbs[-1].setGeometry(QtCore.QRect(432, 11, 51, 32))
        self.dir_pbs[-1].setObjectName(f"pbFolder_{len(self.dir_pbs) - 1}")
        self.dir_pbs[-1].clicked.connect(partial(self._open_folder_dialog, len(self.dir_pbs) - 1))
        self.dir_namelines.append(QtWidgets.QLineEdit(self.dir_groupboxes[-1]))
        self.dir_namelines[-1].setGeometry(QtCore.QRect(550, 15, 101, 21))
        self.dir_namelines[-1].setObjectName(f"leName_{len(self.dir_namelines) - 1}")
        self.dir_namelabels.append(QtWidgets.QLabel(self.dir_groupboxes[-1]))
        self.dir_namelabels[-1].setGeometry(QtCore.QRect(499, 10, 131, 31))
        self.dir_namelabels[-1].setObjectName(f"name_label_{len(self.dir_namelabels) - 1}")
        self.GroupTabBox.addTab(self.dir_groupboxes[-1], f"{len(self.dir_namelabels)}")
        self.dir_folderlines[-1].raise_()
        self.dir_namelines[-1].raise_()
        self.dir_folderlines[-1].setText(self.dirs[-1])
        self.dir_namelines[-1].setText(self.group_labels[-1])
        self.retranslateUi(Dialog, overwrite_inputs=False)

        return True

    def _remove_tab(self, Dialog):
        if len(self.dirs) == 1:
            return False

        index = self.GroupTabBox.currentIndex()
        self.dirs.pop(index)
        self.group_labels.pop(index)
        self.dir_groupboxes.pop(index)
        self.dir_labels.pop(index)
        self.dir_folderlines.pop(index)
        self.dir_pbs.pop(index)
        self.dir_namelines.pop(index)
        self.dir_namelabels.pop(index)
        self.GroupTabBox.removeTab(index)
        for this_tab in range(index, len(self.dirs)):  # rename and reconnect all tabs to the right
            self.GroupTabBox.setTabText(this_tab, f"{this_tab+1}")
            self.dir_pbs[this_tab].clicked.disconnect()
            self.dir_pbs[this_tab].clicked.connect(partial(self._open_folder_dialog, this_tab))
        self.retranslateUi(Dialog, overwrite_inputs=False)

        return True

    def _open_results_folder_dialog(self) -> bool:
        start_folder = self.dir_folderlines[0].text() if self.dir_folderlines[
                                                             0].text() != '' else BatchMeasureDialog.results_folder
        folder = RUtils.open_folder_dialog(f"Results folder", start_folder)

        if folder == '' or folder is False:
            return False

        self.leFolderResults.setText(folder)

        return True


"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import sys
from typing import Optional

import numpy
import pandas as pd
from PyQt6 import QtWidgets
import skimage.filters as skif

from pyjamas.rimage.rimutils import rimutils
from pyjamas.rutils import RUtils
import pyjamas.dialogs.measurepoly as measurepoly
import pyjamas.rannotations.rpolyline as rpolyline
from pyjamas.rcallbacks.rcallback import RCallback


class RCBMeasure(RCallback):
    GAUSSIAN_SIGMA: float = 10.0

    def cbMeasurePoly(self, firstSlice: Optional[int] = None, lastSlice: Optional[int] = None,
                      measurements: Optional[dict] = None, filename: Optional[str] = None) -> dict:
        """
        Measure polylines.

        A dialog will be opened if any parameters are set to None.

        :param firstSlice: slice number for the first slice to use (minimum is 1).
        :param lastSlice: slice number for the last slice to use.
        :param measurements: dictionary with the following keys:

            ``area``:
                True|False
            ``perimeter``:
                True|False
            ``pixels``:
                True|False
            ``image``:
                True|False
            ``sample``:
                True|False
        :param filename: path and file name where the measurement results will be saved; results are saved in .csv format. If filename=='', no results are saved.
        :return: dictionary with measurement results.
        """

        theresults = {}

        # Create and open dialog for measuring polygons.
        if filename is False or filename is None or \
                firstSlice is False or firstSlice is None or lastSlice is False or lastSlice is None:
            # Create a measurement dialog that allows input of all this at once (unless all the parameters are given as arguments).
            dialog = QtWidgets.QDialog()
            ui = measurepoly.MeasurePolyDialog()

            firstSlice = self.pjs.curslice + 1
            lastSlice = 1 if self.pjs.n_frames == 1 else self.pjs.slices.shape[0]
            ui.setupUi(dialog, savepath=self.pjs.cwd, firstslice=firstSlice, lastslice=lastSlice)
            dialog.exec()
            dialog.show()
            # If the dialog was closed by pressing OK, then run the measurements.
            continue_flag = dialog.result() == QtWidgets.QDialog.DialogCode.Accepted
            themeasurements = ui.measurements()
            dialog.close()

        else:
            if firstSlice > lastSlice:
                firstSlice, lastSlice = lastSlice, firstSlice

            if filename not in ('', -1):
                filename = RUtils.set_extension(filename, '.csv')

            if measurements is False or measurements is None:
                themeasurements = {
                    'path': filename,
                    'first': firstSlice,
                    'last': lastSlice,
                    'area': True,
                    'perimeter': True,
                    'pixels': True,
                    'image': True,
                    'sample': False
                }

            else:
                themeasurements = {
                    'path': filename,
                    'first': firstSlice,
                    'last': lastSlice,
                    'area': measurements.get('area', False),
                    'perimeter': measurements.get('perimeter', False),
                    'pixels': measurements.get('pixels', False),
                    'image': measurements.get('image', False),
                    'sample': measurements.get('sample', False)
                }

            continue_flag = True

        if continue_flag:
            theslicenumbers = numpy.arange(themeasurements['first'] - 1, themeasurements['last'])

            theresults = self.measurePolygons(themeasurements, theslicenumbers)

            # If a file name was entered, save the data.
            if themeasurements["path"] not in ('', -1):
                # RUtils.write_dict_csv(filename, theresults)
                # results_df = pd.DataFrame(theresults)
                theresults.to_csv(themeasurements['path'])
                self.pjs.statusbar.showMessage(f'Saved measurements to {themeasurements.get("path")}.')

            elif themeasurements["path"] == '':
                with pd.option_context('display.max_columns', sys.maxsize):
                    print(theresults)
                self.pjs.statusbar.showMessage(f'Measurement results displayed in terminal.')

            elif themeasurements["path"] == -1:
                self.pjs.statusbar.showMessage(f'Measurements completed.')

        return theresults

    def measurePolygons(self, measurements: dict, slices: numpy.ndarray) -> pd.DataFrame:
        # todo: add other measurements: heterogeneity, shape factor, edge-to-centre distance profile, etc.
        # todo: change lists from dictionary into numpy.ndarrays.

        # Returns a pandas DataFrame in which columns represent time points and rows correspond to image statistics or polylines.
        # :param measurements:
        # :param slices: slice indexes -start at 0-, not numbers -start at 1-.
        # :return:

        # Create dictionary with results.
        n_image_metrics: int = 3
        n_polyline_metrics: int = 6
        # Find the maximum polyline id.
        max_n_polylines = max(max(self.pjs.polyline_ids))

        row_names = ['slice_number', 'image_mean', 'image_mode']
        row_names.extend(['area_' + str(i) for i in range(1, max_n_polylines + 1)])
        row_names.extend(['perimeter_' + str(i) for i in range(1, max_n_polylines + 1)])
        row_names.extend(['pixel_values_perimeter_' + str(i) for i in range(1, max_n_polylines + 1)])
        row_names.extend(['pixel_values_interior_' + str(i) for i in range(1, max_n_polylines + 1)])
        row_names.extend(['std_perimeter_' + str(i) for i in range(1, max_n_polylines + 1)])
        row_names.extend(['std_interior_' + str(i) for i in range(1, max_n_polylines + 1)])

        rows: int = n_image_metrics + n_polyline_metrics * max_n_polylines
        columns: int = slices.shape[0]

        measurement_df: pd.DataFrame = pd.DataFrame(numpy.nan * numpy.zeros((rows, columns)), columns=slices + 1,
                                                    index=row_names)

        # If supposed to use a background file, load it ...
        if measurements.get('pixels') and not measurements.get('image') and not measurements.get('sample'):
            bgfilepath = RUtils.set_extension(self.pjs.filename, self.pjs.backgroundimage_extension)
            if os.path.exists(bgfilepath):
                bgstack = rimutils.read_stack(bgfilepath)
            else:
                self.pjs.statusbar.showMessage("Background file not found, reverting to image mean and mode.")
                measurements['image'] = True

        # For every slice ...
        for i in slices:
            # When indexing a DataFrame, the i+1 is the name of the column, not an index.
            measurement_df.loc['slice_number', i + 1] = i + 1

            theimage = self.pjs.slices[i]

            # Find the polylines in this slice.
            polygon_slice = self.pjs.polylines[i]

            # For every polyline ...
            for thepolyid in range(1, max_n_polylines + 1):
                try:
                    thepolyindex = self.pjs.polyline_ids[i].index(thepolyid)
                except ValueError:
                    continue

                # Create a polyline and measure it:
                thepolyline = rpolyline.RPolyline(polygon_slice[thepolyindex])

                # Areas.
                if measurements.get('area'):
                    # Create a polyline and calculate the area.
                    measurement_df.loc['area_' + str(thepolyid), i + 1] = thepolyline.area()

                # Perimeters.
                if measurements.get('perimeter'):
                    measurement_df.loc['perimeter_' + str(thepolyid), i + 1] = thepolyline.perimeter()

                # Pixel values.
                if measurements.get('pixels'):
                    intensities = thepolyline.pixel_values(theimage, self.pjs.brush_size)
                    measurement_df.loc['pixel_values_perimeter_' + str(thepolyid), i + 1] = intensities[0]
                    measurement_df.loc['pixel_values_interior_' + str(thepolyid), i + 1] = intensities[1]
                    measurement_df.loc['std_perimeter_' + str(thepolyid), i + 1] = intensities[2]
                    measurement_df.loc['std_interior_' + str(thepolyid), i + 1] = intensities[3]

            # Image statistics.
            if measurements.get('image'):
                measurement_df.loc['image_mean', i + 1] = numpy.mean(theimage)

                themode = rimutils.mode(theimage)
                measurement_df.loc['image_mode', i + 1] = themode

            if measurements.get('sample'):
                mask = skif.gaussian(numpy.asarray(theimage, dtype=float), RCBMeasure.GAUSSIAN_SIGMA) > numpy.median(
                    theimage)  # also tested the mode here, but the results are very noisy.
                measurement_df.loc['image_mean', i + 1] = numpy.mean(theimage[mask])

                themode = rimutils.mode(theimage[mask])
                measurement_df.loc['image_mode', i + 1] = themode

            # If file-based background subtraction ...
            if measurements.get('pixels') and not measurements.get('image') and not measurements.get('sample'):
                # if the background files has as many slices as the image, use the corresponding slice
                if bgstack.ndim == 3 and bgstack.shape[0] == self.pjs.slices.shape[0]:
                    measurement_df.loc['image_mean', i + 1] = numpy.mean(bgstack[i])

                    themode = rimutils.mode(bgstack[i])
                    measurement_df.loc['image_mode', i + 1] = themode

                # if the number of slices does not match, use the first slice for the background file
                elif bgstack.ndim == 3:
                    measurement_df.loc['image_mean', i + 1] = numpy.mean(bgstack[0])

                    themode = rimutils.mode(bgstack[0])
                    measurement_df.loc['image_mode', i + 1] = themode

                # if the background file has a single slice, use that
                elif bgstack.ndim == 2:
                    measurement_df.loc['image_mean', i + 1] = numpy.mean(bgstack)

                    themode = rimutils.mode(bgstack)
                    measurement_df.loc['image_mode', i + 1] = themode

        return measurement_df



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Optional

from PyQt6 import QtCore, QtWidgets

from pyjamas.rutils import RUtils


class BatchResizeDialog(object):
    input_folder: str = None
    save_folder: str = None
    recurr: bool = None
    im_size: tuple = None

    def __init__(self):
        super().__init__()

    def setupUi(self, Dialog, parameters: Optional[dict] = None):

        if (parameters is None or parameters is False) and BatchResizeDialog.input_folder is None:
            parameters = {'input_folder': '',
                          'save_folder': '',
                          'recurr': True,
                          'im_size': (512, 512)}

        if parameters is not None and parameters is not False:
            BatchResizeDialog.input_folder = parameters.get('input_folder')
            BatchResizeDialog.save_folder = parameters.get('save_folder')
            BatchResizeDialog.recurr = parameters.get('recurr')
            BatchResizeDialog.im_size = parameters.get('im_size')

        Dialog.setObjectName("Dialog")
        Dialog.resize(450, 180)
        self.buttonsOkCancel = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonsOkCancel.setGeometry(QtCore.QRect(140, 140, 200, 32))
        self.buttonsOkCancel.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonsOkCancel.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonsOkCancel.setObjectName("buttonsOkCancel")

        self.label = QtWidgets.QLabel(Dialog)
        self.label.setGeometry(QtCore.QRect(20, 15, 131, 16))
        self.label.setObjectName("label")
        
        self.label_2 = QtWidgets.QLabel(Dialog)
        self.label_2.setGeometry(QtCore.QRect(20, 50, 131, 16))
        self.label_2.setObjectName("label_2")
       
        self.leInputFolder = QtWidgets.QLineEdit(Dialog)
        self.leInputFolder.setGeometry(QtCore.QRect(110, 15, 271, 21))
        self.leInputFolder.setObjectName("leInputFolder")
        
        self.leSaveFolder = QtWidgets.QLineEdit(Dialog)
        self.leSaveFolder.setGeometry(QtCore.QRect(110, 50, 271, 21))
        self.leSaveFolder.setObjectName("leSaveFolder")
        
        self.pbInputFolder = QtWidgets.QPushButton(Dialog)
        self.pbInputFolder.setGeometry(QtCore.QRect(390, 10, 51, 32))
        self.pbInputFolder.setObjectName("pbInputFolder")
        self.pbInputFolder.clicked.connect(self._open_input_folder_dialog)
        
        self.pbSaveFolder = QtWidgets.QPushButton(Dialog)
        self.pbSaveFolder.setGeometry(QtCore.QRect(390, 45, 51, 32))
        self.pbSaveFolder.setObjectName("pbSaveFolder")
        self.pbSaveFolder.clicked.connect(self._open_save_folder_dialog)
    
        self.cbRecurrFlag = QtWidgets.QCheckBox(Dialog)
        self.cbRecurrFlag.setGeometry(QtCore.QRect(20, 110, 150, 20))
        self.cbRecurrFlag.setObjectName("cbRecurrFlag")

        self.label_3 = QtWidgets.QLabel(Dialog)
        self.label_3.setGeometry(QtCore.QRect(20, 80, 190, 20))
        self.label_3.setObjectName("label_3")
        self.leImSize = QtWidgets.QLineEdit(Dialog)
        self.leImSize.setGeometry(QtCore.QRect(210, 80, 71, 21))
        self.leImSize.setObjectName("leImSize")

        self.buttonsOkCancel.raise_()
        self.label.raise_()
        self.label_2.raise_()
        self.leInputFolder.raise_()
        self.leSaveFolder.raise_()
        self.pbInputFolder.raise_()
        self.pbSaveFolder.raise_()
        self.cbRecurrFlag.raise_()
        self.leImSize.raise_()

        self.retranslateUi(Dialog)
        self.buttonsOkCancel.accepted.connect(Dialog.accept)
        self.buttonsOkCancel.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Batch Resize"))
        self.label.setText(_translate("Dialog", "input folder"))
        self.leInputFolder.setText(BatchResizeDialog.input_folder)
        self.label_2.setText(_translate("Dialog", "save folder"))
        self.leSaveFolder.setText(BatchResizeDialog.save_folder)
        self.pbInputFolder.setText(_translate("Dialog", "..."))
        self.pbSaveFolder.setText(_translate("Dialog", "..."))
        self.cbRecurrFlag.setText(_translate("Dialog", "search in subfolders"))
        self.cbRecurrFlag.setChecked(BatchResizeDialog.recurr)
        self.label_3.setText(_translate("Dialog", "new image size (height, width)"))
        self.leImSize.setText(str(BatchResizeDialog.im_size))

    def parameters(self) -> dict:
        BatchResizeDialog.input_folder: str = self.leInputFolder.text()
        BatchResizeDialog.save_folder: str = self.leSaveFolder.text()
        BatchResizeDialog.recurr: bool = self.cbRecurrFlag.isChecked()
        try:
            BatchResizeDialog.im_size = eval(self.leImSize.text())
        except:
            print(f"Incorrect input tuple.")
            BatchResizeDialog.im_size = (512, 512)

        return {
            'input_folder': BatchResizeDialog.input_folder,
            'save_folder': BatchResizeDialog.save_folder,
            'recurr': BatchResizeDialog.recurr,
            'im_size': BatchResizeDialog.im_size
        }

    def _open_input_folder_dialog(self) -> bool:
        start_folder = self.leInputFolder.text() if self.leInputFolder.text() != '' else BatchResizeDialog.input_folder
        folder = RUtils.open_folder_dialog(f"Input folder", start_folder)

        if folder == '' or folder is False:
            return False

        self.leInputFolder.setText(folder)

        return True

    def _open_save_folder_dialog(self) -> bool:
        start_folder = self.leSaveFolder.text() if self.leSaveFolder.text() != '' else BatchResizeDialog.save_folder
        folder = RUtils.open_folder_dialog(f"Save folder", start_folder)

        if folder == '' or folder is False:
            return False

        self.leSaveFolder.setText(folder)

        return True




"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# tests for image menu
# for coverage, run:
# coverage run -m pytest
# or if you want to include branches:
# coverage run --branch -m pytest
# followed by:
# coverage report -i

import gzip
import pickle

import numpy
import os
import pytest
import skimage

from pyjamas.pjscore import PyJAMAS
from pyjamas.rimage.rimutils import rimutils
from pyjamas.rutils import RUtils
from pyjamas.rcallbacks.rcbimage import projection_types
import pyjamas.tests.unit.pjsfixtures as pjsfixtures

PyJAMAS_FIXTURE: PyJAMAS = PyJAMAS()
PyJAMAS_FIXTURE.io.cbLoadTimeSeries(os.path.join(pjsfixtures.FIXTURE_DIR, pjsfixtures.LOGO_FIXTURE))
PyJAMAS_FIXTURE.options.cbSetCWD(pjsfixtures.TMP_DIR)


@pytest.mark.usefixtures("contrastlimits_fixture")
def test_cbAdjustContrast(contrastlimits_fixture):
    PyJAMAS_FIXTURE.image.cbAdjustContrast(contrastlimits_fixture[0], contrastlimits_fixture[1])

    theqimage = PyJAMAS_FIXTURE._imageItem.pixmap().toImage()
    displayed_image = numpy.empty((PyJAMAS_FIXTURE.height, PyJAMAS_FIXTURE.width), dtype=numpy.uint8)

    for row_index in range(PyJAMAS_FIXTURE.height):
        for col_index in range(PyJAMAS_FIXTURE.width):
            displayed_image[row_index, col_index] = theqimage.pixel(col_index, row_index)

    assert numpy.array_equal(displayed_image,
                             numpy.array(
                                 rimutils.stretch(PyJAMAS_FIXTURE.imagedata,
                                                  contrastlimits_fixture[0],
                                                  contrastlimits_fixture[1]),
                                 dtype=numpy.uint8))


def test_cbPlay():
    assert PyJAMAS_FIXTURE.image.cbPlay()


@pytest.mark.usefixtures("rotatecwpath_fixture")
def test_cbRotateImage(rotatecwpath_fixture):
    original_image = PyJAMAS_FIXTURE.slices.copy()

    PyJAMAS_FIXTURE.image.cbRotateImage(PyJAMAS_FIXTURE.image.CW)
    assert numpy.array_equal(PyJAMAS_FIXTURE.slices, skimage.io.imread(rotatecwpath_fixture))

    PyJAMAS_FIXTURE.image.cbRotateImage(PyJAMAS_FIXTURE.image.CCW)
    assert numpy.array_equal(PyJAMAS_FIXTURE.slices, original_image)


@pytest.mark.usefixtures("fliplrpath_fixture", "flipudpath_fixture")
def test_cbFlipImage(fliplrpath_fixture, flipudpath_fixture):
    original_image = PyJAMAS_FIXTURE.slices.copy()

    PyJAMAS_FIXTURE.image.cbFlipImage(PyJAMAS_FIXTURE.image.LEFT_RIGHT)
    assert numpy.array_equal(PyJAMAS_FIXTURE.slices, skimage.io.imread(fliplrpath_fixture))

    PyJAMAS_FIXTURE.image.cbFlipImage(PyJAMAS_FIXTURE.image.LEFT_RIGHT)
    assert numpy.array_equal(PyJAMAS_FIXTURE.slices, original_image)

    PyJAMAS_FIXTURE.image.cbFlipImage(PyJAMAS_FIXTURE.image.UP_DOWN)
    assert numpy.array_equal(PyJAMAS_FIXTURE.slices, skimage.io.imread(flipudpath_fixture))

    PyJAMAS_FIXTURE.image.cbFlipImage(PyJAMAS_FIXTURE.image.UP_DOWN)
    assert numpy.array_equal(PyJAMAS_FIXTURE.slices, original_image)


@pytest.mark.usefixtures("invertpath_fixture")
def test_cbInvertImage(invertpath_fixture):
    PyJAMAS_FIXTURE.image.cbInvertImage()
    theimage = PyJAMAS_FIXTURE.slices.copy()
    PyJAMAS_FIXTURE.options.cbUndo()

    assert numpy.array_equal(theimage, skimage.io.imread(invertpath_fixture))


@pytest.mark.usefixtures("gaussiansigma_fixture")
def test_cbGaussianImage(gaussiansigma_fixture):
    PyJAMAS_FIXTURE.image.cbGaussianImage(gaussiansigma_fixture)
    theimage = PyJAMAS_FIXTURE.slices.copy()
    PyJAMAS_FIXTURE.options.cbUndo()

    assert numpy.array_equal(theimage, rimutils.gaussian_smoothing(PyJAMAS_FIXTURE.slices, gaussiansigma_fixture))


@pytest.mark.usefixtures("gradientpath_fixture")
def test_cbGradientImage(gradientpath_fixture):
    PyJAMAS_FIXTURE.image.cbGradientImage()
    theimage = PyJAMAS_FIXTURE.slices.copy()
    PyJAMAS_FIXTURE.options.cbUndo()

    assert numpy.array_equal(theimage, skimage.io.imread(gradientpath_fixture))


@pytest.mark.usefixtures("projectionpath_fixture")
def test_cbProjectImage(projectionpath_fixture):
    for aprojectionpath, aprojectiontype in zip(projectionpath_fixture, projection_types):
        PyJAMAS_FIXTURE.image.cbProjectImage([], aprojectiontype)
        theproj = PyJAMAS_FIXTURE.slices.copy()
        PyJAMAS_FIXTURE.options.cbUndo()

        assert numpy.array_equal(theproj, skimage.io.imread(aprojectionpath))


@pytest.mark.usefixtures("rescaletuple_fixture")
def test_cbRescaleImage(rescaletuple_fixture):
    PyJAMAS_FIXTURE.image.cbRescaleImage(rescaletuple_fixture)
    theimage = PyJAMAS_FIXTURE.slices.copy()
    PyJAMAS_FIXTURE.options.cbUndo()

    assert numpy.array_equal(theimage, rimutils.rescale(PyJAMAS_FIXTURE.slices, rescaletuple_fixture))


@pytest.mark.usefixtures("registrationannotationspath_fixture")
@pytest.mark.usefixtures("registeredpath_fixture")
def test_cbRegisterImage(registrationannotationspath_fixture, registeredpath_fixture):
    PyJAMAS_FIXTURE.io.cbLoadAnnotations([registrationannotationspath_fixture])
    PyJAMAS_FIXTURE.image.cbRegisterImage()
    theimage = PyJAMAS_FIXTURE.slices.copy()
    PyJAMAS_FIXTURE.options.cbUndo()

    assert numpy.array_equal(theimage, skimage.io.imread(registeredpath_fixture))


def test_cbCropImage():
    theimage = PyJAMAS_FIXTURE.slices.copy()

    # crop around the entire image.
    PyJAMAS_FIXTURE.image.cbCrop(numpy.asarray(
        [[0, 0], [PyJAMAS_FIXTURE.width, 0], [PyJAMAS_FIXTURE.width, PyJAMAS_FIXTURE.height],
         [0, PyJAMAS_FIXTURE.height], [0, 0]]))

    assert numpy.array_equal(theimage, PyJAMAS_FIXTURE.slices)


@pytest.mark.usefixtures("logoannotationspath_fixture")
@pytest.mark.usefixtures("kymographpath_fixture")
def test_cbKymograph(logoannotationspath_fixture, kymographpath_fixture):
    PyJAMAS_FIXTURE.io.cbLoadAnnotations([logoannotationspath_fixture])

    PyJAMAS_FIXTURE.image.cbKymograph(RUtils.qpolygonf2ndarray(PyJAMAS_FIXTURE.polylines[0][0]))
    theimage = PyJAMAS_FIXTURE.slices.copy()

    PyJAMAS_FIXTURE.options.cbUndo()

    assert numpy.array_equal(theimage, skimage.io.imread(kymographpath_fixture))


@pytest.mark.usefixtures("zprojectionspath_fixture")
def test_cbOrthogonalViews(zprojectionspath_fixture):
    PyJAMAS_FIXTURE.image.cbOrthogonalViews()

    assert PyJAMAS_FIXTURE.orthogonal_views and PyJAMAS_FIXTURE.slicetracker

    zx, zy = PyJAMAS_FIXTURE.orthogonal_views.zx.copy(), PyJAMAS_FIXTURE.orthogonal_views.zy.copy()

    PyJAMAS_FIXTURE.image.cbOrthogonalViews()

    assert numpy.array_equal(zx, skimage.io.imread(zprojectionspath_fixture[0]))
    assert numpy.array_equal(zy, skimage.io.imread(zprojectionspath_fixture[1]))

    assert (not PyJAMAS_FIXTURE.orthogonal_views) and (not PyJAMAS_FIXTURE.slicetracker)


def test_cbZoom():
    for izoom in range(len(PyJAMAS_FIXTURE.zoom_factors)):
        PyJAMAS_FIXTURE.image.cbZoom(izoom)
        assert PyJAMAS_FIXTURE.zoom_index == izoom


def test_cbNextFrame():
    theimage = PyJAMAS_FIXTURE.slices[PyJAMAS_FIXTURE.curslice + 1]
    PyJAMAS_FIXTURE.image.cbNextFrame()

    assert numpy.array_equal(theimage, PyJAMAS_FIXTURE.imagedata)


# this should always be after test_cbNextFrame
def test_cbPrevFrame():
    theimage = PyJAMAS_FIXTURE.slices[PyJAMAS_FIXTURE.curslice - 1]
    PyJAMAS_FIXTURE.image.cbPrevFrame()

    assert numpy.array_equal(theimage, PyJAMAS_FIXTURE.imagedata)


def test_cbGoTo():
    theimage = PyJAMAS_FIXTURE.slices[-1]
    PyJAMAS_FIXTURE.image.cbGoTo(PyJAMAS_FIXTURE.n_frames - 1)
    assert numpy.array_equal(theimage, PyJAMAS_FIXTURE.imagedata)

    theimage = PyJAMAS_FIXTURE.slices[0]
    PyJAMAS_FIXTURE.image.cbGoTo(0)
    assert numpy.array_equal(theimage, PyJAMAS_FIXTURE.imagedata)


@pytest.mark.usefixtures("imagepath_fixture")
@pytest.mark.usefixtures("pjsannotationspath_fixture")
@pytest.mark.usefixtures("findseeds_parameters")
@pytest.mark.usefixtures("findseedsannotationspath_fixture")
def test_cbFindSeeds(imagepath_fixture, pjsannotationspath_fixture, findseeds_parameters,
                     findseedsannotationspath_fixture):
    PyJAMAS_FIXTURE.io.cbLoadTimeSeries(imagepath_fixture)
    PyJAMAS_FIXTURE.image.cbFindSeeds(findseeds_parameters[0], findseeds_parameters[1], findseeds_parameters[2],
                                      findseeds_parameters[3], findseeds_parameters[4], findseeds_parameters[5],
                                      preview=False,
                                      wait_for_thread=True)

    with gzip.open(findseedsannotationspath_fixture, "rb") as fh:
        fiducials = pickle.load(fh)

    assert numpy.array_equal(fiducials[0], PyJAMAS_FIXTURE.fiducials[0])


@pytest.mark.usefixtures("imagepath_fixture")
@pytest.mark.usefixtures("propagateseeds0annotationspath_fixture")
@pytest.mark.usefixtures("propagateseeds_parameters")
@pytest.mark.usefixtures("propagateseedsannotationspath_fixture")
def test_cbPropagateSeeds(imagepath_fixture, propagateseeds0annotationspath_fixture, propagateseeds_parameters,
                          propagateseedsannotationspath_fixture):
    PyJAMAS_FIXTURE.io.cbLoadTimeSeries(imagepath_fixture)
    PyJAMAS_FIXTURE.io.cbLoadAnnotations([propagateseeds0annotationspath_fixture])
    PyJAMAS_FIXTURE.image.cbPropagateSeeds(propagateseeds_parameters[0], propagateseeds_parameters[1],
                                           propagateseeds_parameters[2], wait_for_thread=True)

    with gzip.open(propagateseedsannotationspath_fixture, "rb") as fh:
        fiducials = pickle.load(fh)

    assert numpy.array_equal(fiducials, PyJAMAS_FIXTURE.fiducials)


@pytest.mark.usefixtures("imagepath_fixture")
@pytest.mark.usefixtures("propagateseeds0annotationspath_fixture")
@pytest.mark.usefixtures("expandseeds_parameters")
@pytest.mark.usefixtures("expandseedsannotationspath_fixture")
def test_cbExpandSeeds(imagepath_fixture, propagateseeds0annotationspath_fixture, expandseeds_parameters,
                       expandseedsannotationspath_fixture):
    PyJAMAS_FIXTURE.io.cbLoadTimeSeries(imagepath_fixture)
    PyJAMAS_FIXTURE.io.cbLoadAnnotations([propagateseeds0annotationspath_fixture])
    PyJAMAS_FIXTURE.image.cbExpandSeeds(expandseeds_parameters[0], expandseeds_parameters[1], expandseeds_parameters[2],
                                        wait_for_thread=True)

    polylines = PyJAMAS_FIXTURE.polylines.copy()
    PyJAMAS_FIXTURE.io.cbLoadAnnotations([expandseedsannotationspath_fixture])

    assert polylines == PyJAMAS_FIXTURE.polylines


@pytest.mark.usefixtures("imagepath_fixture")
@pytest.mark.usefixtures("propagateseeds0annotationspath_fixture")
@pytest.mark.usefixtures("expandnpropagateseeds_parameters")
@pytest.mark.usefixtures("expandnpropagateseedsannotationspath_fixture")
def test_cbExpandNPropagateSeeds(imagepath_fixture, propagateseeds0annotationspath_fixture, expandnpropagateseeds_parameters,
                       expandnpropagateseedsannotationspath_fixture):
    PyJAMAS_FIXTURE.io.cbLoadTimeSeries(imagepath_fixture)
    PyJAMAS_FIXTURE.io.cbLoadAnnotations([propagateseeds0annotationspath_fixture])
    PyJAMAS_FIXTURE.image.cbExpandNPropagateSeeds(expandnpropagateseeds_parameters[0],
                                                  expandnpropagateseeds_parameters[1],
                                                  expandnpropagateseeds_parameters[2],
                                                  expandnpropagateseeds_parameters[3],
                                                  wait_for_thread=True)

    polylines = PyJAMAS_FIXTURE.polylines.copy()
    PyJAMAS_FIXTURE.io.cbLoadAnnotations([expandnpropagateseedsannotationspath_fixture])

    assert polylines == PyJAMAS_FIXTURE.polylines


def test_cbDisplayInfo():
    assert PyJAMAS_FIXTURE.image.cbDisplayInfo()


PyJAMAS_FIXTURE.app.quit()



"""
    PyJAMAS is Just A More Awesome Siesta
    Copyright (C) 2018  Rodrigo Fernandez-Gonzalez (rodrigo.fernandez.gonzalez@utoronto.ca)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Optional

import numpy
from PyQt6 import QtGui, QtCore

from pyjamas.pjsthreads import ThreadSignals
from pyjamas.rimage.rimml.batchml import BatchML
from pyjamas.rimage.rimml.rimrecurrentneuralnet import rimrecurrentneuralnet
from pyjamas.rimage.rimutils import rimutils
from pyjamas.rutils import RUtils


class BatchRecurrentNeuralNet(BatchML):
    def __init__(self, n_frames: int, classifier: Optional[rimrecurrentneuralnet] = None):
        super().__init__(n_frames, classifier)
        self.prob_arrays = [numpy.empty((1, 0)) for i in range(self.n_frames)]
        self.object_arrays = [numpy.empty((1, 0)) for i in range(self.n_frames)]
        self.object_ids = [numpy.empty((1, 0)) for i in range(self.n_frames)]

    def fit(self, stop_signal: Optional[ThreadSignals] = None) -> bool:
        if self.image_classifier is None:
            return False

        if stop_signal is not None:
            stop_signal.emit('Launching classifier training ...')

        self.image_classifier.fit()

        if stop_signal is not None:
            stop_signal.emit('Classifier training completed.')
        return True

    def predict(self, slices: numpy.ndarray, indices: numpy.ndarray, progress_signal: Optional[ThreadSignals] = None,
                annotations: Optional[numpy.ndarray] = None, ids: Optional[list] = None, max_id: Optional[int] = 0,
                track: bool = False) \
            -> bool:

        # Make sure that the slices are in a 1D numpy array.
        indices = numpy.atleast_1d(indices)
        num_slices = len(indices)

        # For every slice ...
        for i, index in enumerate(indices):
            if slices.ndim > 2:
                theimage = slices[index].copy()
            elif slices.ndim == 2 and index == 0:
                theimage = slices.copy()

            if annotations is not None:
                if not track:
                    if index == 0:  # No previous mask to input
                        themask = numpy.zeros(theimage.shape)
                    elif i == 0:  # Use the polyline annotations made by the user
                        thepolyline = annotations[index - 1].copy()
                        try:
                            themask = rimutils.mask_from_polylines(theimage.shape, thepolyline)
                        except IndexError:
                            themask = numpy.zeros(theimage.shape)
                    else:  # Use the previous classifier output
                        thepolygons = []
                        for apoly in self.object_arrays[index - 1]:
                            thispolygon = QtGui.QPolygonF()
                            for thepoint in apoly:
                                thispolygon.append(QtCore.QPointF(thepoint[0], thepoint[1]))
                            thepolygons.append(thispolygon)
                        try:
                            themask = rimutils.mask_from_polylines(theimage.shape, thepolygons)
                        except IndexError:
                            themask = numpy.zeros(theimage.shape)
                    if numpy.sum(themask) == 0:
                        themask = numpy.ones(theimage.shape)
                    self.object_arrays[index], self.prob_arrays[index] = self.image_classifier.predict(theimage,
                                                                                                       themask)
                    self.object_ids[index] = None  # No reason to believe segmentations are tracked from previous frame

                else:
                    print(f"Slice index {index}, slice number {i}")
                    if index == 0:  # No previous mask to input
                        thepolygons = []
                    elif i == 0:  # Use polyline annotation made by user
                        thepolygons = annotations[index - 1]
                        self.object_ids[index - 1] = ids[index - 1]
                    else:  # Use the previous classifier output
                        thepolygons = []
                        for apoly in self.object_arrays[index - 1]:
                            thispolygon = QtGui.QPolygonF()
                            for thepoint in apoly:
                                thispolygon.append(QtCore.QPointF(thepoint[0], thepoint[1]))
                            thepolygons.append(thispolygon)

                    self.object_ids[index] = []
                    self.object_arrays[index] = []
                    self.prob_arrays[index] = []
                    for obj_num, this_object in enumerate(thepolygons):
                        print(obj_num)
                        try:
                            this_mask = rimutils.mask_from_polylines(imsize=theimage.shape,
                                                                     polylines=[this_object],
                                                                     brushsz=0)
                        except IndexError:  # Polyline not big enough to make a mask from, use empty mask instead
                            this_mask = numpy.ones(theimage.shape, dtype=int)

                        this_object_array, this_prob_array = self.image_classifier.predict(theimage, this_mask)
                        if len(this_object_array) >= 1:  # Object detected
                            for new_obj_num, this_new_object in enumerate(this_object_array):
                                self.object_arrays[index].append(this_new_object)
                                self.prob_arrays[index].append(this_prob_array[new_obj_num])
                                if new_obj_num == 0 and ids is not None:  # Give same ID as prev cell
                                    self.object_ids[index].append(self.object_ids[index - 1][obj_num])
                                else:
                                    self.object_ids[index].append(max_id + 1)
                                    max_id += 1

            else:
                self.object_arrays[index], self.prob_arrays[index] = self.image_classifier.predict(theimage, slice_num=index)
                self.object_ids[index] = None

            if progress_signal is not None:
                progress_signal.emit(int((100 * (i + 1)) / num_slices))

        return True

    def reset_objects(self, n_frames: int) -> bool:
        self.n_frames = n_frames
        self.object_arrays = [numpy.empty((1, 0)) for i in range(self.n_frames)]
        self.prob_arrays = [numpy.empty((1, 0)) for i in range(self.n_frames)]
        self.object_ids = [numpy.empty((1, 0)) for i in range(self.n_frames)]

        return True
